"""
Read NIST MSP formats.

There are an uncountable number of dialects of MSP. This library attempts to
cover a representative subset from the NIST and some found in the wild. If you
encounter an MSP file that does not parse properly, please report it.
"""

import re
import io
import os
import math
import logging
import itertools
import warnings

from dataclasses import dataclass, field
from fractions import Fraction

from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    List,
    Mapping,
    Optional,
    Set,
    Tuple,
    Iterable,
    DefaultDict,
)

from pyteomics import proforma

from mzspeclib import annotation

from mzspeclib.analyte import (
    FIRST_ANALYTE_KEY,
    FIRST_INTERPRETATION_KEY,
    Analyte,
    Interpretation,
    InterpretationMember,
    ProteinDescription,
)

from mzspeclib import const
from mzspeclib.const import (
    CHARGE_STATE,
    CONSENSUS_SPECTRUM,
    MOLECULAR_FORMULA,
    MOLECULAR_MASS,
    PROTON,
    PROFORMA_ION,
    PROFORMA_SEQ,
    SCAN_NUMBER,
    SINGLETON_SPECTRUM,
    SOURCE_FILE,
    SPECTRUM_AGGREGATION_TYPE,
    STRIPPED_PEPTIDE_SEQ as STRIPPED_PEPTIDE_TERM,
    PEAK_ATTRIB,
    PEAK_OBSERVATION_FREQ,
    SELECTED_ION_MZ,
    PRECURSOR_MZ,
    THEORETICAL_MZ,
    CUSTOM_ATTRIBUTE_NAME,
    CUSTOM_ATTRIBUTE_VALUE,
    THEORETICAL_MASS,
    INTENSITY_OF_HIGH_UNASSIGNED_PEAK,
    TOP_20_UNASSIGNED_INTENSITY_FRACTION,
    TOTAL_UNASSIGNED_INTENSITY_FRACTION,
    NUM_UNASSIGNED_PEAKS_IN_TOP_20,
    NUM_PEAKS,
)
from mzspeclib.spectrum import Spectrum, SPECTRUM_NAME
from mzspeclib.attributes import Attribute, AttributeManager, AttributeSet, Attributed

from .base import (
    DEFAULT_VERSION,
    FORMAT_VERSION,
    _PlainTextSpectralLibraryBackendBase,
    LIBRARY_NAME,
    AttributeSetTypes,
    SpectralLibraryBackendBase,
    SpectralLibraryWriterBase,
)
from .utils import try_cast, open_stream, CaseInsensitiveDict


logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

# TODO: Name: could be Compound or SpectrumName
leader_terms = {
    "Name": SPECTRUM_NAME,
    "NAME": SPECTRUM_NAME,
    "Compound": SPECTRUM_NAME,
    "COMPOUND": SPECTRUM_NAME,
}


def _generate_numpeaks_keys():
    w1 = "num"
    w2 = "peaks"
    seps = (" ", "")
    cases = (str.lower, str.title, str.upper)
    w1_cases = [c(w1) for c in cases]
    w2_cases = [c(w2) for c in cases]
    return {
        (w1c + sep + w2c)
        for (sep, w1c, w2c) in itertools.product(seps, w1_cases, w2_cases)
    }


NUM_PEAKS_KEYS = _generate_numpeaks_keys()

LEADER_TERMS_PATTERN = re.compile(r"(Name|NAME|Compound|COMPOUND)\s*:")
LEADER_TERMS_LINE_PATTERN = re.compile(r"(?:Name|NAME|Compound|COMPOUND)\s*:\s+(.+)")

SPACE_SPLITTER = re.compile(r"\s+")

PEPTIDE_MODIFICATION_TERM = "MS:1001471|peptide modification details"


RawPeakLine = Tuple[float, float, str, str]
PeakLine = Tuple[float, float, List[annotation.IonAnnotationBase], List[Any]]

PeakAggregateParseFn = Callable[[str], Any]


class _AttributeHelperBase:
    def add_value(self, key: str, value: Any, container: Attributed):
        """
        Add a single attribute to an :class:`~.Attributed` container.

        Explicitly check for whether the attribute is already present with
        the same value and skip the operation to avoid duplication.
        """
        if container.has_attribute(key):
            existing = container.get_attribute(key)
            if existing == value:
                return
        container.add_attribute(key, value)

    def add_group(self, keys: List[str], values: List[Any], container: Attributed):
        """
        Add a collection of attributes to an :class:`~.Attributed` container
        that will be part of a single group.
        """
        group_id = container.get_next_group_identifier()
        for k, v in zip(keys, values):
            container.add_attribute(k, v, group_id)

    def deduplicate_attribute(self, container: Attributed, attribute_name: str) -> int:
        try:
            attrs = container.get_attribute(attribute_name, raw=True)
        except KeyError:
            return 0
        if isinstance(attrs, list):
            # Sometimes there are multiple entries that map to molecular mass, and one is the nominal
            # mass. Prefer listing the most precise mass first. Apply the same logic to all attributes.
            attrs.sort(key=lambda x: len(str(x.value)), reverse=True)
            container.remove_attribute(attrs[0].key)
            seen_values = set()
            for attr in attrs:
                key = (attr.value, attr.group_id)
                if key in seen_values:
                    continue
                seen_values.add(key)
                container.add_attribute(attr.key, attr.value, attr.group_id)
            return len(attrs) - len(seen_values)
        return 0


class AttributeHandler(_AttributeHelperBase):
    """
    A base type for managing conversion of specific keys or sets of keys and their values
    to controlled vocabulary attributes.
    """

    keys: Collection[str]

    def __init__(self, keys: Collection[str]):
        if isinstance(keys, str):
            keys = (keys,)
        self.keys = keys

    def __contains__(self, key):
        return key in self.keys

    def handle(self, key: str, value: Any, container: Attributed) -> bool:
        """
        Determine what, if any, and how to add this attribute to an
        :class:`~.Attributed` container.

        .. note::
            If this operation is successful, return :const:`True`, otherwise
            return :const:`False` to signal that another :class:`AttributeHandler`
            should try to handle this input, like "bubbling" in event handling
            systems.

        The :meth:`__call__` syntax is short-hand for this method.

        Arguments
        ---------
        key : :class:`str`
            The MSP comment key to map to an attribute
        value : :class:`Any`
            The interpreted value of the MSP comment. It may be a :class:`str` or
            already coerced into another value like a number.
        container : :class:`~.Attributed`
            The object to add the translated attribute to if appropriate.

        Returns
        -------
        :class:`bool`
        """
        raise NotImplementedError()

    def __call__(self, key: str, value: Any, container: Attributed) -> bool:
        """See :meth:`handle`"""
        return self.handle(key, value, container)

    def chain(self, handler: "AttributeHandler") -> "AttributeHandlerChain":
        """
        Combine this :class:`AttributeHandler` with another :class:`AttributeHandler`
        to be applied in sequence.

        The :meth:`__and__` syntax is short-hand for this method.

        Returns
        -------
        :class:`AttributeHandlerChain`
        """
        return AttributeHandlerChain([self, handler])

    def __and__(self, handler: "AttributeHandler") -> "AttributeHandlerChain":
        return self.chain(handler)


class MappingAttributeHandler(AttributeHandler):
    """
    An :class:`AttributeHandler` which uses a :class:`Mapping` to conveniently
    define the relationship between a comment key and the appropriate controlled
    vocabulary term.

    The behavior is handled differently depending upon the value in ``self.keys[key]``:
        - Scalar: ``self.keys[key]`` is the controlled vocabulary term used,
          and the comment value is minimally processed.
        - List of 2 Items and comment value is :const:`None`: ``self.keys[key][0]`` is the controlled vocabulary term
          and ``self.keys[key][1]`` is the value.
        - :class:`AttributeHandler`: All handling is delegated to ``self.keys[key](key, value, container)``
        - Mapping: The comment value is used to find the appropriate scalar or list to add.
        - List of Pairs: All pairs from the list are added with :meth:`add_group`.

    More complex transformations should be defined using an explicit rule
    like :class:`FunctionAttributeHandler` or some other sub-class if appropriate.
    """

    keys: Dict[str, Any]

    def __init__(self, keys: Dict[str, Any]):
        self.keys = keys

    def handle(self, key: str, value: Any, container: Attributed) -> bool:
        trans_key = self.keys[key]
        if value is None:
            if isinstance(trans_key, list):
                k, v = trans_key
                v = try_cast(v)
                self.add_value(k, v, container)
            else:
                return False
        elif isinstance(trans_key, str):
            self.add_value(trans_key, try_cast(value), container)
        elif isinstance(trans_key, dict):
            if value in trans_key:
                # If the mapping is a plain string, add it
                if isinstance(trans_key[value], str):
                    key, value = trans_key[value].split("=")
                    self.add_value(key, try_cast(value), container)
                # Or if it is a list, then there are multiple terms to add within a group
                elif isinstance(trans_key[value], list):
                    if len(trans_key[value]) == 1:
                        for item in trans_key[value]:
                            self.add_value(item[0], try_cast(item[1]), container)
                    else:
                        k, v = zip(*trans_key[value])
                        self.add_group(k, map(try_cast, v), container)
            else:
                return False
        elif isinstance(trans_key, AttributeHandler):
            return trans_key(key, value, container)
        return True


class AttributeHandlerChain:
    """
    An :class:`AttributeHandler` that sequentially applies a list
    of :class:`AttributeHandler`s until it finds one that works.

    This type exists to make handler composition easier.
    """

    chain: List[AttributeHandler]

    def __init__(self, chain: List[AttributeHandler]):
        self.chain = chain

    def handle(self, key: str, value: Any, container: Attributed) -> bool:
        result = True
        for handler in self.chain:
            result |= handler.handle(key, value, container)
            if not result:
                return result
        return result

    def __contains__(self, key):
        for handler in self.chain:
            if key in handler:
                return True
        return False

    def __iter__(self):
        return iter(self.chain)

    def __call__(self, key: str, value: Any, container: Attributed) -> bool:
        return self.handle(key, value, container)

    def chain(self, handler: "AttributeHandler") -> "AttributeHandlerChain":
        return self.__class__(self.chain + [handler])

    def add(self, handler: AttributeHandler):
        self.chain.append(handler)
        return handler


class FunctionAttributeHandler(AttributeHandler):
    func: Callable[[str, Any, Attributed], bool]

    def __init__(
        self, keys: Collection[str], func: Callable[[str, Any, Attributed], bool]
    ):
        super().__init__(keys)
        self.func = func

    def handle(self, key: str, value: Any, container: Attributed) -> bool:
        return self.func(key, value, container)

    @classmethod
    def wraps(cls, *keys):
        def wrapper(func):
            return cls(keys, func)

        return wrapper


class DispatchingAttributeHandler(AttributeHandlerChain):
    mapping: Dict[str, AttributeHandler]

    def __init__(self, chain: List[AttributeHandler] = None):
        if not chain:
            chain = []
        super().__init__(chain)
        self.mapping = CaseInsensitiveDict()
        for handler in self:
            for key in handler.keys:
                self.mapping[key] = handler

    def handle(self, key: str, value: Any, container: Attributed) -> bool:
        handler = self.mapping[key]
        return handler(key, value, container)

    def __contains__(self, key):
        return key in self.mapping

    def add(self, handler: AttributeHandler):
        super().add(handler)
        for key in handler.keys:
            self.mapping[key] = handler
        return handler


spectrum_terms = CaseInsensitiveDict(
    {
        "Charge": CHARGE_STATE,
        "precursor_charge": CHARGE_STATE,
        "precursorcharge": CHARGE_STATE,
        "Parent": SELECTED_ION_MZ,
        "PrecursorMonoisoMZ": PRECURSOR_MZ,
        "ObservedPrecursorMZ": PRECURSOR_MZ,
        "PrecursorMZ": PRECURSOR_MZ,
        "PRECURSORMZ": PRECURSOR_MZ,
        "precursor": PRECURSOR_MZ,
        "precursor_mass": PRECURSOR_MZ,
        "precursormass": PRECURSOR_MZ,
        "Mz_exact": PRECURSOR_MZ,
    }
)

analyte_terms = CaseInsensitiveDict(
    {
        # "MW": "MS:1000224|molecular mass",
        # "total exact mass": "MS:1000224|molecular mass",
        # "ExactMass": "MS:1000224|molecular mass",
        # "exact_mass": "MS:1000224|molecular mass",
        # "exact mass": "MS:1000224|molecular mass",
        "molecular formula": MOLECULAR_FORMULA,
        "Formula": MOLECULAR_FORMULA,
        "formula": MOLECULAR_FORMULA,
        "SMILES": "MS:1000868|SMILES formula",
        "InChIKey": "MS:1002894|InChIKey",
        "InChI": "MS:1003403|InChI",
        "Precursor_type": "MS:1002813|adduct ion formula",
        # "Theo_mz_diff": "MS:1003209|monoisotopic m/z deviation",
        "Scan": {
            "Mods": PEPTIDE_MODIFICATION_TERM,
            "Naa": "MS:1003043|number of residues",
        },
        "Pep": {
            "Tryptic": [
                ["MS:1003048|number of enzymatic termini", 2],
                ["MS:1001045|cleavage agent name", "MS:1001251|Trypsin"],
            ],
            "SemiTryptic": [
                ["MS:1003048|number of enzymatic termini", 1],
                ["MS:1001045|cleavage agent name", "MS:1001251|Trypsin"],
            ],
            "N-Semitryptic": [
                ["MS:1003048|number of enzymatic termini", 1],
                ["MS:1001045|cleavage agent name", "MS:1001251|Trypsin"],
            ],
            "C-Semitryptic": [
                ["MS:1003048|number of enzymatic termini", 1],
                ["MS:1001045|cleavage agent name", "MS:1001251|Trypsin"],
            ],
            "Tryptic/miss_good_confirmed": [
                ["MS:1003048|number of enzymatic termini", 2],
                ["MS:1003044|number of missed cleavages", "0"],
                ["MS:1001045|cleavage agent name", "MS:1001251|Trypsin"],
            ],
            "Tryptic/miss_bad_confirmed": [
                ["MS:1003048|number of enzymatic termini", 2],
                ["MS:1003044|number of missed cleavages", ">0"],
                ["MS:1001045|cleavage agent name", "MS:1001251|Trypsin"],
            ],
        },
        "MC": "MS:1003044|number of missed cleavages",
        "NMC": "MS:1003044|number of missed cleavages",
        "Mods": PEPTIDE_MODIFICATION_TERM,
        "Naa": "MS:1003043|number of residues",
        "Mz_av": "MS:1003054|theoretical average m/z",
    }
)


_HCD = [
    const.DISSOCIATION_METHOD,
    const.HCD,
]

# TODO: qtof -> CAD
instrument_dispatch = CaseInsensitiveDict(
    {
        "it": [
            [
                const.DISSOCIATION_METHOD,
                const.TRAP_CID,
            ]
        ],
        "hcd": [_HCD],
        "QExactive": [_HCD],
        "Elite": [_HCD],
    }
)


other_terms = CaseInsensitiveDict(
    {
        "Single": [
            const.SPECTRUM_AGGREGATION_TYPE,
            const.SINGLETON_SPECTRUM,
        ],
        "Consensus": [
            const.SPECTRUM_AGGREGATION_TYPE,
            const.CONSENSUS_SPECTRUM,
        ],
        "Inst": instrument_dispatch,
        "Instrument_type": instrument_dispatch,
        "Spec": {
            "Consensus": [
                [
                    const.SPECTRUM_AGGREGATION_TYPE,
                    const.CONSENSUS_SPECTRUM,
                ]
            ]
        },
        "Scan": SCAN_NUMBER,
        "Origfile": SOURCE_FILE,
        "filename": SOURCE_FILE,
        "file_name": SOURCE_FILE,
        "Sample": const.SAMPLE_NAME,
        "Filter": const.FILTER_STRING,
        "FTResolution": "MS:1000028|detector resolution",
        "ms1PrecursorAb": "MS:1003085|previous MSn-1 scan precursor intensity",
        "Precursor1MaxAb": const.PRECURSOR_APEX_INTENSITY,
        "Purity": const.ISOLATION_WINDOW_PRECURSOR_PURITY,
        "Num peaks": NUM_PEAKS,
        "Num Peaks": NUM_PEAKS,
        "num_peaks": NUM_PEAKS,
        "numpeaks": NUM_PEAKS,
        "Run": SOURCE_FILE,
        "Splash": "MS:1002599|splash key",
    }
)


@FunctionAttributeHandler.wraps("num_unassigned_peaks")
def unassigned_peaks_handler(key: str, value: str, container: Attributed) -> bool:
    is_top_20 = False
    if isinstance(value, str):
        if "/" in value:
            if "/20" in value:
                is_top_20 = True
            value = value.split("/")[0]
        value = int(value)
    assert isinstance(value, int)
    if is_top_20:
        container.add_attribute(
            NUM_UNASSIGNED_PEAKS_IN_TOP_20, value
        )
    else:
        container.add_attribute("MS:1003288|number of unassigned peaks", value)
    return True


interpretation_terms = CaseInsensitiveDict(
    {
        "Unassigned_all_20ppm": TOTAL_UNASSIGNED_INTENSITY_FRACTION,
        "Unassign_all": TOTAL_UNASSIGNED_INTENSITY_FRACTION,
        "top_20_num_unassigned_peaks_20ppm": unassigned_peaks_handler,
        "num_unassigned_peaks_20ppm": unassigned_peaks_handler,
        "num_unassigned_peaks": unassigned_peaks_handler,
        "max_unassigned_ab_20ppm": INTENSITY_OF_HIGH_UNASSIGNED_PEAK,
        "max_unassigned_ab": INTENSITY_OF_HIGH_UNASSIGNED_PEAK,
        "Unassigned_20ppm": TOP_20_UNASSIGNED_INTENSITY_FRACTION,
        "Unassigned": TOP_20_UNASSIGNED_INTENSITY_FRACTION,
    }
)


interpretation_member_terms = CaseInsensitiveDict(
    {
        "Q-value": const.Q_VALUE,
    }
)


species_map = {
    "human": [
        [const.TAXONOMY_NCBI_TAX_ID, "NCBITaxon:9606|Homo sapiens"],
        [const.TAXONOMY_SCIENTIFIC_NAME, "Homo sapiens"],
        [const.TAXONOMY_COMMON_NAME, "human"],
    ],
    "zebrafish": [
        [const.TAXONOMY_NCBI_TAX_ID, "NCBITaxon:7955|Danio rerio"],
        [const.TAXONOMY_SCIENTIFIC_NAME, "Danio rerio"],
        [const.TAXONOMY_COMMON_NAME, "zebra fish"],
    ],
    "chicken": [
        [const.TAXONOMY_NCBI_TAX_ID, "NCBITaxon:9031|Gallus gallus"],
        [const.TAXONOMY_SCIENTIFIC_NAME, "Gallus gallus"],
        [const.TAXONOMY_COMMON_NAME, "chicken"],
    ],
    "rat": [
        [const.TAXONOMY_NCBI_TAX_ID, "NCBITaxon:10116|Rattus norvegicus"],
        [const.TAXONOMY_SCIENTIFIC_NAME, "Rattus norvegicus"],
        [const.TAXONOMY_COMMON_NAME, "rat"],
    ],
}


MODIFICATION_NAME_MAP = {
    "CAM": "Carbamidomethyl",
    "Pyro_glu": "Glu->pyro-Glu",  # Resolves UNIMOD ambiguity
    "Pyro-glu": "Gln->pyro-Glu",
    "Oxidation": "Oxidation",
    "Phospho": "Phospho",
    "TMT6plex": "TMT6plex",
    "iTRAQ": "iTRAQ",
    "Acetyl": "Acetyl",
    "TMT": "TMT6plex",
}

TERMINAL_MODIFICATIONS = {"Acetyl", "TMT6plex"}


# TODO: ppm is unsigned, add mass calculation to determine true mass accuracy

annotation_pattern = re.compile(
    r"""^
(?:(?:(?P<series>[abyxcz]\.?)(?P<ordinal>\d+))|
   (:?(?P<series_internal>(?:Int/[ARNDCEQGHKMFPSTWYVILJarndceqghkmfpstwyvilj]+)|(?:[mM](?P<internal_start>\d+):(?P<internal_end>\d+))))|
   (?P<precursor>p)|
   (:?I(?P<immonium>[ARNDCEQGHKMFPSTWYVIL])(?:(?P<immonium_modification>CAM)|[A-Z])?)|
   (?P<reporter>(?P<reporter_label>TMT|ITRAQ|iTRAQ)(?P<reporter_mass>\d+[NC]?))|
   (?:_(?P<external_ion>[^\s,/]+))|
   (?P<unannotated>\?)
)
(?P<neutral_losses>((?:[+-]\d*[A-Z][A-Za-z0-9]*)|(?:[+-]iTRAQ|TMT)|(?:[+-]\d+?(?!i)))+)?
(?:(?:\[M(?P<adducts>(:?[+-]\d*[A-Z][A-Za-z0-9]*)+)\])?
(?:\^(?P<charge>[+-]?\d+))?
(?:(?P<isotope>[+-]?\d*)i)?)+
(?:/(?:Dev=)?(?P<mass_error>[+-]?\d+(?:\.\d+))(?P<mass_error_unit>ppm)?)?
""",
    re.X,
)


class MSPAnnotationStringParser(annotation.AnnotationStringParser):
    def _dispatch(
        self,
        annotation_string,
        data,
        adducts,
        charge,
        isotope,
        neutral_losses,
        analyte_reference,
        mass_error,
        confidence,
        **kwargs,
    ):
        data["sequence_ordinal"] = None
        data["sequence_internal"] = None
        return super()._dispatch(
            annotation_string,
            data,
            adducts,
            charge,
            isotope,
            neutral_losses,
            analyte_reference,
            mass_error,
            confidence,
            **kwargs,
        )

    def _dispatch_internal_peptide_fragment(
        self,
        data: Dict[str, Any],
        adducts: List,
        charge: int,
        isotope: int,
        neutral_losses: List,
        analyte_reference: Any,
        mass_error: Any,
        **kwargs,
    ):
        if data["internal_start"]:
            # The hybrid pattern detected an m<start>:<end> type string, not an Int/seq string
            return super(
                MSPAnnotationStringParser, self
            )._dispatch_internal_peptide_fragment(
                data,
                adducts,
                charge,
                isotope,
                neutral_losses,
                analyte_reference,
                mass_error,
                **kwargs,
            )

        spectrum = kwargs.get("spectrum")
        if spectrum is None:
            raise ValueError(
                "Cannot infer sequence coordinates from MSP internal fragmentation notation without"
                " a reference to the spectrum, please pass spectrum as a keyword argument"
            )
        sequence = self._get_peptide_sequence_for_analyte(spectrum, analyte_reference)
        subseq = data["series_internal"]
        if subseq.startswith("Int/"):
            subseq = subseq[4:]
        subseq = subseq.upper()
        try:
            start_index = sequence.index(subseq)
        except ValueError as err:
            raise ValueError(
                f"Cannot locate internal subsequence {subseq} in {sequence}"
            ) from err
        end_index = start_index + len(subseq)
        data["internal_start"] = start_index + 1
        data["internal_end"] = end_index
        return super(
            MSPAnnotationStringParser, self
        )._dispatch_internal_peptide_fragment(
            data,
            adducts,
            charge,
            isotope,
            neutral_losses,
            analyte_reference,
            mass_error,
            **kwargs,
        )

    def _coerce_isotope(self, data):
        value = data.get("isotope")
        if value is not None:
            if value == "":
                data["isotope"] = "1"
        return super()._coerce_isotope(data)

    def _coerce_neutral_losses(self, data: Dict[str, str]) -> List:
        return super()._coerce_neutral_losses(data)

    def _coerce_analyte_reference(self, data: Dict[str, str]) -> str:
        return None

    def _dispatch_immonium(
        self,
        data: Dict[str, Any],
        adducts: List,
        charge: int,
        isotope: int,
        neutral_losses: List,
        analyte_reference: Any,
        mass_error: Any,
        **kwargs,
    ):
        modification = data["immonium_modification"]
        if modification is not None:
            try:
                modification = MODIFICATION_NAME_MAP[modification]
                data["immonium_modification"] = modification
            except KeyError as err:
                print(f"Failed to convert immonium ion modification {modification}")
        return super(MSPAnnotationStringParser, self)._dispatch_immonium(
            data,
            adducts,
            charge,
            isotope,
            neutral_losses,
            analyte_reference,
            mass_error,
            **kwargs,
        )

    def _get_peptide_sequence_for_analyte(
        self, spectrum: Spectrum, analyte_reference: Optional[Any] = None
    ) -> str:
        if analyte_reference is None:
            if len(spectrum.analytes) == 0:
                return None
            else:
                analyte_reference = spectrum.analytes[FIRST_ANALYTE_KEY].id
        analyte = spectrum.analytes.get(analyte_reference)
        if analyte is None:
            return None
        return analyte.get_attribute("MS:1000888|stripped peptide sequence")


parse_annotation = MSPAnnotationStringParser(annotation_pattern)
MODIFICATION_LIST_PARSER = re.compile(
    r"(\d+),([ARNDCEQGHKMFPSTWYVIL_\-]),([A-Za-z0-9_\-]+)"
)


class ModificationParser:
    pattern: re.Pattern
    modification_map: Dict[str, str]
    unknown_modifications: Set[str]

    def __init__(self, pattern: str = None, modification_map: Dict[str, str] = None):
        if pattern is None:
            pattern = MODIFICATION_LIST_PARSER.pattern
        if modification_map is None:
            modification_map = dict(MODIFICATION_NAME_MAP)
        self.pattern = re.compile(pattern)
        self.modification_map = modification_map
        self.unknown_modifications = set()

    def __call__(self, text: str):
        return self.parse(text)

    def parse(self, text: str) -> List[Tuple[int, str, str]]:
        if not isinstance(text, str) or not text:
            return []
        i = 0
        n = len(text)

        mods = []
        while i < n and text[i].isdigit():
            i += 1

        for position, residue, mod in self.pattern.findall(text):
            position = int(position)
            if mod not in self.modification_map:
                warnings.warn(
                    f"{mod} is not found in the known MSP modification mapping. Using this name verbatim"
                )
                modification_name = mod
                self.modification_map[mod] = mod
                self.unknown_modifications.add(mod)
            else:
                modification_name = self.modification_map[mod]
            mods.append((position, residue, modification_name))
        return mods


def null_handler(key: str, value: str, container: Attributed) -> bool:
    return True


msp_spectrum_attribute_handler = DispatchingAttributeHandler()
msp_analyte_attribute_handler = DispatchingAttributeHandler()

msp_interpretation_attribute_handler = DispatchingAttributeHandler()
msp_interpretation_attribute_handler.add(MappingAttributeHandler(interpretation_terms))

msp_spectrum_attribute_handler.add(FunctionAttributeHandler("Nprot", null_handler))
msp_spectrum_attribute_handler.add(FunctionAttributeHandler("Peptype", null_handler))

msp_spectrum_attribute_handler.add(MappingAttributeHandler(spectrum_terms))


@msp_spectrum_attribute_handler.add
@FunctionAttributeHandler.wraps("Spectrum_type")
def ms_level_handler(key: str, value: str, container: Attributed) -> bool:

    attr_name = const.MS_LEVEL
    if value is None:
        return False
    if isinstance(value, str):
        if value.lower().startswith("ms"):
            value = int(value[2:])
    container.add_attribute(attr_name, value)
    return True


@msp_spectrum_attribute_handler.add
@FunctionAttributeHandler.wraps(
    "ionmode", "ion_mode", "ionization mode", "IONMODE", "Ion_mode", "MS_mode"
)
def polarity_handler(key: str, value: str, container: Attributed) -> bool:
    polarity_term = const.SCAN_POLARITY
    positive = const.POSITIVE_SCAN
    negative = const.NEGATIVE_SCAN


    if isinstance(value, str):
        value = value.lower().strip("\"'")
        if value == "positive" or value == "+":
            value = positive
        elif value == "negative" or value == "-":
            value = negative
        else:
            raise ValueError(f"Can't infer scan polarity from {value}")
        container.add_attribute(polarity_term, value)
        return True
    return False


@msp_spectrum_attribute_handler.add
@FunctionAttributeHandler.wraps("HCD")
def dissociation_method_handler(key: str, value: str, container: Attributed) -> bool:
    container.add_attribute(
        const.DISSOCIATION_METHOD,
        const.HCD,
    )
    found_match = False
    if value is not None:
        match = re.match(r"([\d\.]+)\s*ev", value, flags=re.IGNORECASE)
        if match is not None:
            found_match = True
            group_identifier = container.get_next_group_identifier()

            container.add_attribute(
                const.COLLISION_ENERGY,
                try_cast(match.group(1)),
                group_identifier,
            )
            container.add_attribute(
                const.UNIT, const.ELECTRONVOLT, group_identifier
            )
        match = re.match(r"([\d\.]+)\s*%", value)
        if match is not None:
            found_match = True
            group_identifier = container.get_next_group_identifier()
            container.add_attribute(
                const.COLLISION_ENERGY,
                try_cast(match.group(1)),
                group_identifier,
            )
            container.add_attribute(const.UNIT, const.PERCENT, group_identifier)
    return found_match


@msp_spectrum_attribute_handler.add
@FunctionAttributeHandler.wraps(
    "Collision_energy", "CE", "colenergy", "collisionenergy", "ionization energy"
)
def collision_energy_handler(key: str, value: str, container: Attributed) -> bool:
    if isinstance(value, str):
        if "NCE" in value:
            return normalized_collision_energy_handler(key, value, container)
        match = re.match(r"([\d\.]+)(:?eV|EV|ev|Ev)?", value.strip('"'))
        if match is not None:
            value = try_cast(match.group(1))
        else:
            warnings.warn(
                f"Failed to parse {value} for {key} in collision_energy_handler"
            )
    if value is not None:
        group_identifier = container.get_next_group_identifier()
        container.add_attribute(const.COLLISION_ENERGY, value, group_identifier)
        container.add_attribute(
            const.UNIT, const.ELECTRONVOLT, group_identifier
        )
        return True
    return False


@msp_spectrum_attribute_handler.add
@FunctionAttributeHandler.wraps("NCE", "nce")
def normalized_collision_energy_handler(
    key: str, value: str, container: Attributed
) -> bool:
    if isinstance(value, str):
        match = re.match(r"([\d\.]+)", value)
        if match is not None:
            value = try_cast(match.group(1))
    if value is not None:
        group_identifier = container.get_next_group_identifier()
        container.add_attribute(
            "MS:1000138|normalized collision energy", value, group_identifier
        )
        container.add_attribute(
            const.UNIT, const.PERCENT, group_identifier
        )
        return True
    return False


@msp_spectrum_attribute_handler.add
@FunctionAttributeHandler.wraps("RT", "rettime", "retentiontime", "rtinseconds")
def rt_handler(key, value, container) -> bool:
    if not isinstance(value, str):
        if key.lower() == "rtinseconds":
            group_identifier = container.get_next_group_identifier()
            container.add_attribute(
                const.RETENTION_TIME,
                try_cast(value),
                group_identifier,
            )
            container.add_attribute(
                const.UNIT, const.SECOND, group_identifier
            )
        else:
            warnings.warn(f"Unable to infer retention time unit from {key!r}")
            container.add_attribute(const.RETENTION_TIME, try_cast(value))
        return True
    else:
        match = re.match(r"([\d\.]+)\s*(\D*)", value)
        if match is not None:
            if match.group(2):
                container.add_attribute(
                    "ERROR", f"Need more RT parsing code to handle this value"
                )
                return False
            else:
                group_identifier = container.get_next_group_identifier()
                container.add_attribute(
                    const.RETENTION_TIME,
                    try_cast(match.group(1)),
                    group_identifier,
                )
                #### If the value is greater than 250, assume it must be seconds
                if float(match.group(1)) > 250 or key.lower() == "rtinseconds":
                    container.add_attribute(
                        const.UNIT, const.SECOND, group_identifier
                    )
                #### Although normally assume minutes
                else:
                    container.add_attribute(
                        const.UNIT, const.MINUTE, group_identifier
                    )
                return True
    return False


@msp_spectrum_attribute_handler.add
@FunctionAttributeHandler.wraps("ms2IsolationWidth")
def isolation_width_handler(key, value, container) -> bool:
    if value is None:
        return False
    group_identifier = container.get_next_group_identifier()
    container.add_attribute(
        "MS:1000828|isolation window lower offset", (float(value) / 2), group_identifier
    )
    container.add_attribute(const.UNIT, const.MZ, group_identifier)
    group_identifier = container.get_next_group_identifier()
    container.add_attribute(
        "MS:1000829|isolation window upper offset", (float(value) / 2), group_identifier
    )
    container.add_attribute(const.UNIT, const.MZ, group_identifier)
    return True


@msp_interpretation_attribute_handler.add
@FunctionAttributeHandler.wraps("Mz_diff", "Theo_mz_diff")
def mz_diff_handler(key, value, container: Attributed) -> bool:
    if isinstance(value, float):
        # We must be dealing with a unit-less entry.
        group_identifier = container.get_next_group_identifier()
        container.add_attribute(const.DELTA_MZ, abs(value), group_identifier)
        container.add_attribute(const.UNIT, const.MZ, group_identifier)
    else:
        match = re.match(r"([\-\+e\d\.]+)\s*ppm", value, flags=re.IGNORECASE)
        if match is not None:
            group_identifier = container.get_next_group_identifier()
            container.add_attribute(
                const.DELTA_MZ, try_cast(match.group(1)), group_identifier
            )
            container.add_attribute(
                const.UNIT, const.PPM, group_identifier
            )
        else:
            match = re.match(r"([\-\+e\d\.]+)\s*", value)
            if match is not None:
                group_identifier = container.get_next_group_identifier()
                container.add_attribute(
                    const.DELTA_MZ, try_cast(match.group(1)), group_identifier
                )
                container.add_attribute(
                    const.UNIT, const.MZ, group_identifier
                )
            else:
                return False
    return True


@msp_interpretation_attribute_handler.add
@FunctionAttributeHandler.wraps("Dev_ppm")
def dev_ppm_handler(key, value, container) -> bool:
    if value is None:
        return False
    group_identifier = container.get_next_group_identifier()
    container.add_attribute(const.DELTA_MZ, try_cast(value), group_identifier)
    container.add_attribute(
        const.UNIT, const.PPM, group_identifier
    )
    return True


@msp_spectrum_attribute_handler.add
@FunctionAttributeHandler.wraps("Nrep", "Nreps")
def nreps_handler(key, value, container):
    if value is None:
        return False
    if not isinstance(value, str):
        value = str(value)
    match = re.match(r"(\d+)/(\d+)", value)
    if match is not None:
        container.add_attribute(
            const.NUMBER_OF_REPLICATE_SPECTRA_USED, try_cast(match.group(1))
        )
        container.add_attribute(
            const.NUMBER_OF_REPLICATE_SPECTRA_AVAILABLE, try_cast(match.group(2))
        )
        return True
    else:
        match = re.match(r"(\d+)", value)
        if match is not None:
            container.add_attribute(
                const.NUMBER_OF_REPLICATE_SPECTRA_USED, try_cast(match.group(1))
            )
            container.add_attribute(
                const.NUMBER_OF_REPLICATE_SPECTRA_AVAILABLE,
                try_cast(match.group(1)),
            )
            return True
        else:
            return False


@msp_analyte_attribute_handler.add
@FunctionAttributeHandler.wraps("Organism")
def organism_handler(key, value, container):
    if value is None:
        return False
    value = value.strip('"')

    if value in species_map:
        group_identifier = container.get_next_group_identifier()
        for item in species_map[value]:
            container.add_attribute(item[0], try_cast(item[1]), group_identifier)
        return True
    return False


@msp_analyte_attribute_handler.add
@FunctionAttributeHandler.wraps("Protein")
def protein_handler(key, value, container: Attributed):
    if value is None:
        return False
    key = "MS:1000885|protein accession"
    match = re.match(r"\(pre=(.),post=(.)\)", value)
    group_identifier = None
    if match is not None:
        value = value[: match.start()]
        group_identifier = container.get_next_group_identifier()
        container.add_attribute(
            "MS:1001112|n-terminal flanking residue",
            match.group(1),
            group_identifier=group_identifier,
        )
        container.add_attribute(
            "MS:1001113|c-terminal flanking residue",
            match.group(2),
            group_identifier=group_identifier,
        )
    container.add_attribute(
        key,
        re.sub(r"\(pre=(.),post=(.)\)", "", value.strip('"').strip("'")),
        group_identifier=group_identifier,
    )
    return True


@msp_spectrum_attribute_handler.add
@FunctionAttributeHandler.wraps("BasePeak")
def base_peak_handler(key, value, container: Attributed):
    if value is None:
        return False
    value = float(value)
    group_id = container.get_next_group_identifier()
    container.add_attribute("MS:1000505|base peak intensity", value, group_id)
    container.add_attribute(
        const.UNIT, "MS:1000131|number of detector counts", group_id
    )
    return True


class _UnknownTermTracker:
    counts: DefaultDict

    def add(self, key: str, value: Optional[str] = None):
        """Add an unknown attribute to the tracker"""
        raise NotImplementedError()

    def items(self):
        return self.counts.items()


class UnknownKeyValueTracker(_UnknownTermTracker):
    """
    A diagnostic tool for tracking attributes with values that the parser doesn't know how to interpret.

    This tracker holds both keys and values, and can grow quite large. For debugging purposes only.
    """

    def __init__(self) -> None:
        self.counts = DefaultDict(lambda: DefaultDict(int))

    def add(self, key: str, value: Optional[str] = None):
        self.counts[key][value] += 1


class UnknownKeyTracker(_UnknownTermTracker):
    """A diagnostic tool for tracking attributes that the parser doesn't know how to interpret."""

    def __init__(self) -> None:
        self.counts = DefaultDict(int)

    def add(self, key: str, value: Optional[str] = None):
        self.counts[key] += 1


protein_attributes_to_group = [
    "MS:1003048|number of enzymatic termini",
    "MS:1001045|cleavage agent name",
    "MS:1001112|n-terminal flanking residue",
    "MS:1001113|c-terminal flanking residue",
    "MS:1003044|number of missed cleavages",
    "MS:1000885|protein accession",
]


@dataclass
class AttributeRuleBase(_AttributeHelperBase):
    engine: Optional["AttributeRulesEngine"] = field(default=None, init=False)

    def is_applicable(
        self,
        attributes: Dict[str, Any],
        spectrum: Spectrum,
        analyte: Analyte,
        interpretation: Interpretation,
        interpretation_member: InterpretationMember,
    ) -> bool:
        raise NotImplementedError()

    def process_attributes(
        self,
        attributes: Dict[str, Any],
        spectrum: Spectrum,
        analyte: Analyte,
        interpretation: Interpretation,
        interpretation_member: InterpretationMember,
        unknown_terms: Optional[List[str]] = None,
    ) -> List[str]:
        raise NotImplementedError()


class StrippedPeptideParserRule(AttributeRuleBase):
    def is_applicable(
        self,
        attributes: Dict[str, Any],
        spectrum: Spectrum,
        analyte: Analyte,
        interpretation: Interpretation,
        interpretation_member: InterpretationMember,
    ) -> bool:
        return not analyte.has_attribute(
            STRIPPED_PEPTIDE_TERM
        ) and spectrum.has_attribute(SPECTRUM_NAME)

    def process_attributes(
        self,
        attributes: Dict[str, Any],
        spectrum: Spectrum,
        analyte: Analyte,
        interpretation: Interpretation,
        interpretation_member: InterpretationMember,
        unknown_terms: Optional[List[str]] = None,
    ) -> List[str]:
        name = spectrum.get_attribute(SPECTRUM_NAME)
        if name:
            match = re.match(r"([ARNDCEQGHKMFPSTWYVIL]+)/(\d+)", name)
            if match:
                self.add_value(STRIPPED_PEPTIDE_TERM, match.group(1), analyte)
                self.add_value(CHARGE_STATE, try_cast(match.group(2)), analyte)
        []


@dataclass
class FullNameParsingRule(AttributeRuleBase):
    attribute: str = "Fullname"

    def is_applicable(
        self,
        attributes: Dict[str, Any],
        spectrum: Spectrum,
        analyte: Analyte,
        interpretation: Interpretation,
        interpretation_member: InterpretationMember,
    ) -> bool:
        return self.attribute in attributes

    def parse_value(self, value: str):
        match = re.match(
            r"([A-Z\-\*])\.([A-Z]+)\.([A-Z\-\*])/*([\d]*)",
            value,
        )
        if match is None:
            return None
        peptide = match.group(2)
        nterm = match.group(1)
        cterm = match.group(3)
        charge = match.group(4)
        return peptide, nterm, cterm, charge

    def process_attributes(
        self,
        attributes: Dict[str, Any],
        spectrum: Spectrum,
        analyte: Analyte,
        interpretation: Interpretation,
        interpretation_member: InterpretationMember,
        unknown_terms: Optional[List[str]] = None,
    ) -> List[str]:
        value = attributes[self.attribute]
        more_unknown_terms = []
        if value is not None:
            match = self.parse_value(value)
            if match is not None:
                (peptide, nterm, cterm, charge) = match
                analyte.add_attribute(STRIPPED_PEPTIDE_TERM, peptide)
                analyte.add_attribute("MS:1001112|n-terminal flanking residue", nterm)
                analyte.add_attribute("MS:1001113|c-terminal flanking residue", cterm)
                if charge:
                    analyte.add_attribute(CHARGE_STATE, try_cast(charge))
                attributes.pop(self.attribute)
            else:
                more_unknown_terms.append(self.attribute)
        else:
            more_unknown_terms.append(self.attribute)
        return more_unknown_terms


@dataclass
class PeptideNotationUnpackingRule(AttributeRuleBase):
    modification_parser: ModificationParser

    def is_applicable(
        self,
        attributes: Dict[str, Any],
        spectrum: Spectrum,
        analyte: Analyte,
        interpretation: Interpretation,
        interpretation_member: InterpretationMember,
    ) -> bool:
        return bool(analyte.attributes)

    def process_attributes(
        self,
        attributes: Dict[str, Any],
        spectrum: Spectrum,
        analyte: Analyte,
        interpretation: Interpretation,
        interpretation_member: InterpretationMember,
        unknown_terms: Optional[List[str]] = None,
    ) -> List[str]:
        self.complete_analyte(analyte, spectrum)
        return []

    def complete_analyte(self, analyte: Analyte, spectrum: Spectrum):
        if analyte.has_attribute(STRIPPED_PEPTIDE_TERM):
            peptide = proforma.ProForma.parse(
                analyte.get_attribute(STRIPPED_PEPTIDE_TERM)
            )

            if analyte.has_attribute(PEPTIDE_MODIFICATION_TERM):
                modification_details = analyte.get_attribute(PEPTIDE_MODIFICATION_TERM)
                mods = self.modification_parser(modification_details)
                for position, residue, mod in mods:
                    if position == 0 and mod in TERMINAL_MODIFICATIONS:
                        peptide.n_term = [proforma.GenericModification(mod)]
                    else:
                        seqpos = list(peptide.sequence[position])
                        if not seqpos[1]:
                            seqpos[1] = [proforma.GenericModification(mod)]
                        peptide.sequence[position] = tuple(seqpos)
                        assert seqpos[0] == residue

            charge = None
            if analyte.charge:
                charge = analyte.charge
            elif spectrum.precursor_charge:
                charge = spectrum.precursor_charge

            if charge is not None:
                peptide.charge_state = charge
                analyte.add_attribute(PROFORMA_ION, str(peptide))
            else:
                analyte.add_attribute(PROFORMA_SEQ, str(peptide))

            if analyte.has_attribute(PEPTIDE_MODIFICATION_TERM):
                analyte.remove_attribute(PEPTIDE_MODIFICATION_TERM)
            analyte.add_attribute(THEORETICAL_MASS, peptide.mass)

        self.pack_protein_description(analyte)

        self.deduplicate_attribute(analyte, MOLECULAR_MASS)
        self.deduplicate_attribute(analyte, MOLECULAR_FORMULA)

    def pack_protein_description(self, analyte: Analyte):
        table = {}
        max_len = 0

        # Collect terms that describe a protein
        for term in protein_attributes_to_group:
            if not analyte.has_attribute(term):
                table[term] = []
                continue
            values = analyte.get_attribute(term, raw=True)
            if not isinstance(values, list):
                values = [values]
            max_len = max(max_len, len(values))
            table[term] = values

        # Ensure that all arrays are the same length
        for k, v in table.items():
            if len(v) < max_len:
                v.extend([None] * (max_len - len(v)))

        # Group together terms and remove the previous entries
        groups = []
        for i in range(max_len):
            group = []
            for k, v in table.items():
                if v[i] is None:
                    continue
                group.append((k, v[i].value))
                try:
                    analyte.remove_attribute(v[i].key, v[i].group_id)
                except KeyError:
                    # Sometimes, there are multiple entries for the same attribute,
                    # some with a group ID, some without. If an entry without a group
                    # ID is removed, every entry is removed, and subsequent removals
                    # do not find the entry.
                    pass
            groups.append(group)

        # Now add back the groups
        for group in groups:
            analyte.add_attribute_group(group)


@dataclass
class AnalyteMassDescriptorDiscriminationRule(AttributeRuleBase):

    SELECTED_ION_MZ_TERM = "MS:1000744|selected ion m/z"
    ADDUCT_ION_MASS_TERM = "MS:1003243|adduct ion mass"
    MOLECULAR_MASS_TERM = "MS:1000224|molecular mass"
    EXP_PREC_MONO_MZ_TERM = "MS:1003208|experimental precursor monoisotopic m/z"
    THEO_AVG_MZ_TERM = "MS:1003054|theoretical average m/z"

    precursor_ion_mz_terms = {
        "PrecursorMonoisoMZ": EXP_PREC_MONO_MZ_TERM,
        "ObservedPrecursorMZ": EXP_PREC_MONO_MZ_TERM,
        "PrecursorMZ": EXP_PREC_MONO_MZ_TERM,
        "PRECURSORMZ": EXP_PREC_MONO_MZ_TERM,
        "precursor": EXP_PREC_MONO_MZ_TERM,
        "precursor_mass": EXP_PREC_MONO_MZ_TERM,
        "precursormass": EXP_PREC_MONO_MZ_TERM,
        "Mz_exact": EXP_PREC_MONO_MZ_TERM,
        "Mz_av": THEO_AVG_MZ_TERM,
        "Parent": SELECTED_ION_MZ_TERM,
    }

    molecular_mass_terms = {
        "MW": ADDUCT_ION_MASS_TERM,
        "total exact mass": MOLECULAR_MASS_TERM,
        "ExactMass": MOLECULAR_MASS_TERM,
        "exact_mass": MOLECULAR_MASS_TERM,
    }

    @dataclass
    class _CommentAttrValue:
        comment: str
        attribute: str
        value: float
        error: Optional[float] = None

    def is_applicable(
        self,
        attributes: Dict[str, Any],
        spectrum: Spectrum,
        analyte: Analyte,
        interpretation: Interpretation,
        interpretation_member: InterpretationMember,
    ) -> bool:
        attributes = CaseInsensitiveDict(attributes)
        for key in itertools.chain.from_iterable(
            (self.molecular_mass_terms, self.precursor_ion_mz_terms)
        ):
            if key in attributes:
                return True
        return False

    def collect_mz_terms(
        self, attributes: Dict[str, Any]
    ) -> List["AnalyteMassDescriptorDiscriminationRule._CommentAttrValue"]:
        terms = []
        for k, v in self.precursor_ion_mz_terms.items():
            if k in attributes:
                terms.append(self._CommentAttrValue(k, v, try_cast(attributes[k])))
        return terms

    def collect_mass_terms(
        self, attributes: Dict[str, Any]
    ) -> List["AnalyteMassDescriptorDiscriminationRule._CommentAttrValue"]:
        terms = []
        for k, v in self.molecular_mass_terms.items():
            if k in attributes:
                terms.append(self._CommentAttrValue(k, v, try_cast(attributes[k])))
        return terms

    def process_attributes(
        self,
        attributes: Dict[str, Any],
        spectrum: Spectrum,
        analyte: Analyte,
        interpretation: Interpretation,
        interpretation_member: InterpretationMember,
        unknown_terms: Optional[List[str]] = None,
    ) -> List[str]:
        peptide = analyte.peptide

        adduct = PROTON
        molecular_mass_computed = None
        adduct_mass_computed = None

        molecular_mass_candidates: List[
            AnalyteMassDescriptorDiscriminationRule._CommentAttrValue
        ] = []
        adduct_mass_candidates: List[
            AnalyteMassDescriptorDiscriminationRule._CommentAttrValue
        ] = []

        explained_attributes = []
        # This entry has a peptide analyte
        if peptide is not None:
            peptide_mass = peptide.mass
            analyte_charge = spectrum.precursor_charge

            mass_terms = self.collect_mass_terms(attributes)

            if analyte_charge is None:
                analyte_charge = spectrum.precursor_charge

            if analyte_charge is not None:
                molecular_mass_computed = peptide_mass
                adduct_mass_computed = peptide_mass + adduct * analyte_charge

            categories = [
                (
                    self.MOLECULAR_MASS_TERM,
                    molecular_mass_computed,
                ),
                (
                    self.ADDUCT_ION_MASS_TERM,
                    adduct_mass_computed,
                ),
            ]

            for attr in mass_terms:
                best_category = None
                best_error = float("inf")
                for kind, theoretical in categories:
                    err = abs(attr.value - theoretical)
                    if err < best_error:
                        best_category = kind
                        best_error = err

                attr.error = best_error
                if best_error > 0.5 and math.isfinite(best_error):
                    logger.warning(
                        "Large error detected in best mass category classification: %s %r",
                        best_category,
                        best_error,
                    )
                if best_category == self.MOLECULAR_MASS_TERM:
                    molecular_mass_candidates.append(attr)
                elif best_category == self.ADDUCT_ION_MASS_TERM:
                    adduct_mass_candidates.append(attr)
                else:
                    logger.warning(
                        "No mass terms detected for spectrum %s", spectrum.key
                    )

            if adduct_mass_candidates:
                adduct_mass_candidates.sort(key=lambda x: x.error)
                if len(adduct_mass_candidates) > 1:
                    warnings.warn(
                        "Multiple adduct mass candidates found: %r",
                        adduct_mass_candidates,
                    )
                term = adduct_mass_candidates[0]
                explained_attributes.append(term.comment)
                analyte.add_attribute(self.ADDUCT_ION_MASS_TERM, term.value)
            if molecular_mass_candidates:
                molecular_mass_candidates.sort(key=lambda x: x.error)
                if len(molecular_mass_candidates) > 1:
                    warnings.warn(
                        "Multiple molecular mass candidates found: %r",
                        molecular_mass_candidates,
                    )
                term = adduct_mass_candidates[0]
                explained_attributes.append(term.comment)
                analyte.add_attribute(self.MOLECULAR_MASS_TERM, term.value)

        # This entry has no molecular analyte or we don't know how to get a mass from it
        else:
            mass_terms = self.collect_mass_terms(attributes)
            for term in mass_terms:
                analyte.add_attribute(term.attribute, term.value)
                explained_attributes.append(term.comment)

        if unknown_terms:
            for term in explained_attributes:
                unknown_terms.remove(term)


@dataclass
class AttributeRulesEngine:
    other_manager: AttributeHandler
    analyte_manager: AttributeHandler
    interpretation_manager: AttributeHandler
    interpretation_member_manager: AttributeHandler
    spectrum_attribute_handler: AttributeHandler
    analyte_attribute_handler: AttributeHandler

    early_rules: List[AttributeRuleBase] = field(default_factory=list)
    late_rules: List[AttributeRuleBase] = field(default_factory=list)

    def bind_rule(self, rule: AttributeRuleBase):
        rule.engine = self
        return rule

    def __post_init__(self):
        for rule in self.early_rules:
            self.bind_rule(rule)
        for rule in self.late_rules:
            self.bind_rule(rule)

    def process_attributes(
        self,
        attributes: Dict[str, Any],
        spectrum: Spectrum,
        analyte: Analyte,
        interpretation: Interpretation,
        interpretation_member: InterpretationMember,
    ):
        unknown_terms = []

        for early_rule in self.early_rules:
            if early_rule.is_applicable(
                attributes, spectrum, analyte, interpretation, interpretation_member
            ):
                early_rule.process_attributes(
                    attributes, spectrum, analyte, interpretation, interpretation_member
                )

        for attribute in attributes:
            # Skip a leader term that we already processed
            if attribute in leader_terms:
                continue
            if not attribute:
                continue
            if attribute in self.other_manager:
                if not self.other_manager(attribute, attributes[attribute], spectrum):
                    unknown_terms.append(attribute)

            elif attribute in self.analyte_manager:
                if not self.analyte_manager(attribute, attributes[attribute], analyte):
                    unknown_terms.append(attribute)

            elif attribute in self.interpretation_manager:
                if not self.interpretation_manager(
                    attribute, attributes[attribute], interpretation
                ):
                    unknown_terms.append(attribute)

            elif attribute in self.interpretation_member_manager:
                if not self.interpretation_member_manager(
                    attribute, attributes[attribute], interpretation_member
                ):
                    unknown_terms.append(attribute)

            elif attribute in self.spectrum_attribute_handler:
                if not self.spectrum_attribute_handler(
                    attribute, attributes[attribute], spectrum
                ):
                    unknown_terms.append(attribute)

            elif attribute in self.analyte_attribute_handler:
                if not self.analyte_attribute_handler(
                    attribute, attributes[attribute], analyte
                ):
                    unknown_terms.append(attribute)
            else:
                unknown_terms.append(attribute)

        for late_rule in self.late_rules:
            if late_rule.is_applicable(
                attributes, spectrum, analyte, interpretation, interpretation_member
            ):
                late_rule.process_attributes(
                    attributes,
                    spectrum,
                    analyte,
                    interpretation,
                    interpretation_member,
                    unknown_terms=unknown_terms,
                )

        return unknown_terms


def proportion_parser(aggregation: str) -> float:
    """Parse for fractions or percentages"""
    if "/" in aggregation:
        aggregation: Fraction = Fraction(aggregation)
        return float(aggregation)
    else:
        return float(aggregation)


@dataclass
class PeakAggregationParser:
    """
    Parse peak aggregation information.

    Subtypes may produce different attributes.

    Attributes
    ----------
    peak_attributes : list[(:class:`~.Attribute`, :class:`PeakAggregateParseFn`)]
        The attributes this parser expects
    """

    peak_attributes: List[Tuple[Attribute, PeakAggregateParseFn]] = field(
        default_factory=lambda: [
            (Attribute(PEAK_ATTRIB, PEAK_OBSERVATION_FREQ), proportion_parser)
        ]
    )

    def __call__(
        self, aggregation: str, wrap_errors: bool = True, **kwargs
    ) -> List[Tuple[Attribute, float]]:
        parsed = []

        if isinstance(aggregation, str):
            aggregation = SPACE_SPLITTER.split(aggregation)

        for (i, token), (k, parser) in zip(
            enumerate(aggregation), self.peak_attributes
        ):
            try:
                result = parser(token)
                parsed.append((k, result))
            except Exception as err:
                if not wrap_errors:
                    raise err from None
                else:
                    logger.error(f"Failed to parse aggregation at {i}")
                    parsed.append((k, token))
        return parsed


@dataclass
class PeakParsingStrategy:
    """
    A combination of peak annotation parsing and peak aggregation parsing
    strategies.

    Attributes
    ----------
    annotation_parser : :class:`MSPAnnotationStringParser` or :const:`None`
        The peak annotation parser for this format.

    aggregation_parser : :class:`PeakAggregationParser` or :const:`None`
        The peak aggregation parser for this format
    """

    annotation_parser: Optional[MSPAnnotationStringParser] = field(
        default=parse_annotation
    )
    aggregation_parser: Optional[PeakAggregationParser] = field(
        default_factory=PeakAggregationParser
    )

    def has_aggregation(self) -> bool:
        return self.aggregation_parser is not None

    def has_annotation(self) -> bool:
        return self.annotation_parser is not None

    def parse_aggregation(self, aggregation: str, wrap_errors: bool = True, **kwargs):
        return self.aggregation_parser(
            aggregation=aggregation, wrap_errors=wrap_errors, **kwargs
        )

    def parse_annotation(self, annotation: str, wrap_errors: bool = True, **kwargs):
        return self.annotation_parser(
            annotation_string=annotation, wrap_errors=wrap_errors, **kwargs
        )

    def process_peak_list(self, spectrum: Spectrum):
        aggregation_metrics_used = DefaultDict(int)
        for i, peak in enumerate(spectrum.peak_list):
            try:
                parsed_interpretation = self.parse_annotation(
                    peak[2], spectrum=spectrum
                )
            except ValueError as err:
                message = str(err)
                raise ValueError(
                    f"An error occurred while parsing the peak annotation for peak {i}: {message}"
                ) from err

            if (
                parsed_interpretation
                and isinstance(parsed_interpretation[0], annotation.InvalidAnnotation)
                and peak[3]
            ):
                logger.debug(
                    "Failed to parse interpretation string %r with assumed %d aggregation fields, trying to parse the combined string",
                    peak[2],
                    len(peak[3]),
                )

                peak[2] = " ".join([peak[2]] + peak[3])
                peak[3] = []
                try:
                    parsed_interpretation = self.parse_annotation(
                        peak[2], spectrum=spectrum
                    )
                except ValueError as err:
                    message = str(err)
                    raise ValueError(
                        f"An error occurred while parsing the peak annotation for peak {i}: {message}"
                    ) from err

            peak[2] = parsed_interpretation
            if peak[3]:
                aggregations = []
                for agg_type, agg in self.parse_aggregation(peak[3]):
                    aggregation_metrics_used[agg_type] += 1
                    aggregations.append(agg)

                peak[3] = aggregations

        aggregation_metrics = [
            metric for metric, count in aggregation_metrics_used.items() if count > 0
        ]

        if aggregation_metrics:
            spectrum.add_attribute_group(aggregation_metrics)

    def parse_peak_list(self, peak_lines: Iterable[str]) -> List[RawPeakLine]:
        peak_list = []
        for values in peak_lines:
            interpretations = ""
            aggregation = ""
            if len(values) == 1:
                mz = values
                intensity = "1"
            if len(values) == 2:
                mz, intensity = values
            elif len(values) == 3:
                mz, intensity, interpretations = values
            elif len(values) > 3:
                mz, intensity, interpretations = values[0:2]
            else:
                mz = "1"
                intensity = "1"

            interpretations = interpretations.strip('"')
            aggregation = None
            if interpretations.startswith("?"):
                parts = SPACE_SPLITTER.split(interpretations)
                if len(parts) > 1:
                    # Some msp files have a concept for ?i, but this requires a definition
                    interpretations = "?"
                    aggregation = parts[1:]
            else:
                if " " in interpretations:
                    parts = SPACE_SPLITTER.split(interpretations)
                    interpretations = parts[0]
                    aggregation = parts[1:]

            #### Add to the peak list
            peak_list.append(
                [float(mz), float(intensity), interpretations, aggregation]
            )
        return peak_list

    def __call__(self, peak_lines: Iterable[str]) -> list:
        raw_peaks = self.parse_peak_list(peak_lines)
        return raw_peaks


class MSPSpectralLibrary(_PlainTextSpectralLibraryBackendBase):
    """
    A reader for the plain text NIST MSP spectral library format.

    The MSP format is only roughly defined, and does places few
    constraints on the meanings of spectrum attributes. This parser
    attempts to cover a variety of different ways that MSPs found
    "in the wild" have denoted different spectrum properties, but
    is neither exhaustive nor nuanced enough to know from context
    exactly what those files' authors intended, making a best guess
    at when they correspond to in the controlled vocabulary mapping
    for :mod:`mzspeclib`

    Attributes
    ----------
    modification_parser : :class:`ModificationParser`
        A parser for peptide modifications
    unknown_attributes : :class:`_UnknownTermTracker`
        A tracker for unknown attributes. Used to tell how much information
        the reader is unable to map onto the controlled vocabulary.
    """

    file_format = "msp"
    format_name = "msp"

    modification_parser: ModificationParser
    unknown_attributes: _UnknownTermTracker

    def __init__(
        self, filename, index_type=None, read_metadata=True, create_index: bool = True
    ):
        super().__init__(filename, index_type, read_metadata, create_index=create_index)
        self.modification_parser = ModificationParser()
        self.unknown_attributes = UnknownKeyTracker()

    @classmethod
    def guess_from_header(cls, filename: str) -> bool:
        with open_stream(filename, "r") as stream:
            first_line = stream.readline()
            if LEADER_TERMS_PATTERN.match(first_line):
                return True
        return False

    def read_header(self) -> bool:
        filename = self.filename
        file_like_object = isinstance(filename, io.IOBase)

        stream = open_stream(filename, "rt")
        match, offset = self._parse_header_from_stream(stream)
        if file_like_object:
            if stream.seekable():
                stream.seek(0)
            stream.detach()
        else:
            stream.close()
        return match

    def _parse_header_from_stream(self, stream: io.IOBase) -> Tuple[bool, int]:
        first_line = stream.readline()
        if isinstance(first_line, bytes):
            first_line = first_line.decode("utf8")
        attributes = AttributeManager()
        attributes.add_attribute(FORMAT_VERSION, DEFAULT_VERSION)
        if isinstance(self.filename, (str, os.PathLike)):
            attributes.add_attribute(
                LIBRARY_NAME, self.filename.rsplit(".msp", 1)[0].split(os.sep)[-1]
            )
        elif hasattr(stream, "name"):
            attributes.add_attribute(
                LIBRARY_NAME, stream.name.rsplit(".msp", 1)[0].split(os.sep)[-1]
            )
        self.attributes.clear()
        self.attributes._from_iterable(attributes)
        if LEADER_TERMS_PATTERN.match(first_line):
            return True, 0
        return False, 0

    def create_index(self) -> int:
        """
        Populate the spectrum index

        Returns
        -------
        n_spectra: int
            The number of entries read
        """
        #### Check that the spectrum library filename isvalid
        filename = self.filename

        begin_loc = None
        file_like_object = False
        #### Determine the filesize
        try:
            file_size = os.path.getsize(filename)
        except TypeError:
            if isinstance(filename, io.IOBase):
                file_like_object = True
                if filename.seekable():
                    begin_loc = filename.tell()
                    filename.seek(0, os.SEEK_END)
                    file_size = filename.tell()
                    filename.seek(0)
        infile = open_stream(filename, "r")

        state = "header"
        spectrum_buffer = []
        n_spectra = 0
        start_index = 0
        file_offset = 0
        line_beginning_file_offset = 0
        spectrum_file_offset = 0
        spectrum_name = ""

        # Required for counting file_offset manually (LF vs CRLF)
        infile.readline()
        file_offset_line_ending = len(infile.newlines) - 1
        infile.seek(0)

        # if debug:
        #     eprint("INFO: Reading..", end='', flush=True)
        logger.debug(f"Reading {filename} ({file_size} bytes)...")
        while 1:
            line = infile.readline()
            if len(line) == 0:
                break

            line_beginning_file_offset = file_offset

            #### tell() is twice as slow as counting it myself
            # file_offset = infile.tell()
            file_offset += len(line) + file_offset_line_ending

            line = line.rstrip()
            # TODO: Name: could be Compound or SpectrumName
            if state == "header":
                if LEADER_TERMS_PATTERN.match(line):
                    state = "body"
                    spectrum_file_offset = line_beginning_file_offset
                else:
                    continue
            if state == "body":
                if len(line) == 0:
                    continue
                if LEADER_TERMS_PATTERN.match(line):
                    if len(spectrum_buffer) > 0:
                        self.index.add(
                            number=n_spectra + start_index,
                            offset=spectrum_file_offset,
                            name=spectrum_name,
                            analyte=None,
                        )
                        n_spectra += 1
                        spectrum_buffer = []
                        #### Commit every now and then
                        if n_spectra % 10000 == 0:
                            self.index.commit()
                            logger.info(
                                f"... Indexed  {file_offset} bytes, {n_spectra} spectra read"
                            )

                    spectrum_file_offset = line_beginning_file_offset
                    spectrum_name = LEADER_TERMS_LINE_PATTERN.match(line).group(1)

                spectrum_buffer.append(line)
        logger.debug(f"Processed {file_offset} bytes, {n_spectra} spectra read")
        self.index.add(
            number=n_spectra + start_index,
            offset=spectrum_file_offset,
            name=spectrum_name,
            analyte=None,
        )
        self.index.commit()
        n_spectra += 1

        #### Flush the index
        self.index.commit()

        if not file_like_object:
            infile.close()
        else:
            infile.detach()
            if begin_loc is not None:
                filename.seek(begin_loc)
        return n_spectra

    def _buffer_from_stream(self, infile: Iterable[str]) -> List[str]:
        state = "body"
        spectrum_buffer = []

        file_offset = 0
        line_beginning_file_offset = 0
        for line in infile:
            line_beginning_file_offset = file_offset
            file_offset += len(line)
            line = line.rstrip()
            if state == "body":
                if len(line) == 0:
                    continue
                if LEADER_TERMS_PATTERN.match(line):
                    if len(spectrum_buffer) > 0:
                        return spectrum_buffer
                spectrum_buffer.append(line)
        return spectrum_buffer

    def _make_peak_parsing_strategy(self) -> PeakParsingStrategy:
        return PeakParsingStrategy(parse_annotation)

    def _parse(self, buffer: Iterable[str], spectrum_index: int = None) -> Spectrum:
        #### Start in the header section of the entry
        in_header = True

        #### Reset all spectrum properties in case there is already data here
        attributes = {}
        peak_list = []
        peak_line_parser: PeakParsingStrategy = self._make_peak_parsing_strategy()

        #### Loop through each line in the buffered list
        for line in buffer:
            #### If in the the header portion of the entry
            if in_header:
                key = value = None
                #### Extract the key,value pair by splitting on the *first* colon with optional whitespace
                match = re.match(
                    r"\s*#", line
                )  # Assume lines starting with # are comments
                if match:
                    continue
                elif line.count(":") > 0:
                    key, value = re.split(r":\s*", line, 1)
                    attributes[key] = value
                elif line.count("=") > 0:
                    key, value = re.split(r"=\s*", line, 1)
                    attributes[key] = value
                elif line.count("\t") > 0:
                    warnings.warn(f"Line {line!r} looks like a peak annotation?")
                    in_header = False
                else:
                    key = line
                    attributes[key] = value

                #### If the key is "Num peaks" then we're done with the header and peak list follows
                if key in NUM_PEAKS_KEYS:
                    in_header = False

                #### The "Comment" key requires special parsing
                if key == "Comment" or key == "Comments":
                    #### Remove it from attributes
                    del attributes[key]
                    self._parse_comment(value, attributes)

            #### Else in the peaks section. Parse the peaks.
            else:
                #### Split into the expected three values
                values = re.split(r"\s+", line, maxsplit=2)
                # Sometimes MSP files have multiple peaks on the same line, delimited by ';',
                # so we must potentially parse more than one peak per line.
                if values[1].endswith(";"):
                    buffered_peaks = []
                    buffered_peaks.append(values[:2])
                    buffered_peaks[0][1] = buffered_peaks[0][1].strip(";")
                    rest = values[0]
                    for block in re.split(r";\s?", rest):
                        if block:
                            buffered_peaks.append(re.split(r"\s+", block, maxsplit=2))
                    peak_list.extend(buffered_peaks)
                else:
                    peak_list.append(values)

        peak_list = peak_line_parser(peak_list)

        #### Now convert the format attributes to standard ones
        spectrum = self._make_spectrum(peak_list, attributes)
        if spectrum_index is not None:
            spectrum.index = spectrum_index
        else:
            spectrum.index = -1
        spectrum.key = spectrum.index + 1
        peak_line_parser.process_peak_list(spectrum)
        return spectrum

    def _parse_annotation(self, annotation: str, wrap_errors: bool = True, **kwargs):
        return parse_annotation(
            annotation_string=annotation, wrap_errors=wrap_errors, **kwargs
        )

    def _parse_comment(self, value: str, attributes: Attributed):
        comment_items = re.split(" ", value)

        #### Any spaces within quotes are then de-split
        fixed_comment_items = []
        quote_counter = 0
        new_item = ""
        for item in comment_items:
            if new_item > "":
                new_item = new_item + " "
            new_item = new_item + item
            n_quotes = new_item.count('"')
            if n_quotes % 2 == 0:
                fixed_comment_items.append(new_item)
                new_item = ""

        #### Try to split each item on the first = character and store in attributes
        for item in fixed_comment_items:
            #### If there is an = then split on the first one and store key and value
            if item.count("=") > 0:
                comment_key, comment_value = item.split("=", 1)
                cleaned_key = comment_key.strip('"')
                if len(cleaned_key) != len(comment_key):
                    comment_value = comment_value.strip('"')
                attributes[cleaned_key] = try_cast(comment_value)

            #### Otherwise just store the key with a null value
            else:
                attributes[item] = None

    def _make_attribute_handlers(self) -> AttributeRulesEngine:
        """
        Create the attribute handling scopes that map this flavor of MSP's
        attributes onto controlled vocabulary terms in context.

        This method should be overridden in sub-classes to allow them
        to change the meanings of attributes, add new ones, or otherwise
        redirect how they are interpreted.

        See the :class:`AttributeHandler` type tree for more details about
        how the distributed predicates are resolved.

        Returns
        -------
        other_manager : :class:`AttributeHandler`
            The attribute handler for uncategorized attributes that will be added
            to a :class:`Spectrum`.
        analyte_manager : :class:`AttributeHandler`
            The attribute handler for attributes that will be added to a :class:`Analyte`
        interpretation_manager : :class:`AttributeHandler`
            The attribute handler for attributes that will be added to a :class:`Interpretation`
        interpretation_member_manager : :class:`AttributeHandler`
            The attribute handler for attributes that will be added to a :class:`InterpretationMember`
        spectrum_manager : :class:`AttributeHandler`
            The attribute handler for attributes that will be added to a :class:`Spectrum`
        analyte_fallback_manager : :class:`AttributeHandler`
            The attribute handler for attributes that will be tried for any attribute
            that fails to be categorized by all of the other managers to be added to the
            :class:`Analyte` before labeling the attribute as "unknown".
        """
        other_manager = MappingAttributeHandler(other_terms)
        analyte_manager = MappingAttributeHandler(analyte_terms)
        interpretation_manager = msp_interpretation_attribute_handler
        interpretation_member_manager = MappingAttributeHandler(
            interpretation_member_terms
        )

        return AttributeRulesEngine(
            other_manager,
            analyte_manager,
            interpretation_manager,
            interpretation_member_manager,
            msp_spectrum_attribute_handler,
            msp_analyte_attribute_handler,
            early_rules=[FullNameParsingRule()],
            late_rules=[
                StrippedPeptideParserRule(),
                PeptideNotationUnpackingRule(self.modification_parser),
                AnalyteMassDescriptorDiscriminationRule(),
            ],
        )

    def _make_spectrum(self, peak_list: List, attributes: Mapping[str, str]):
        spectrum = self._new_spectrum()
        interpretation = self._new_interpretation(FIRST_INTERPRETATION_KEY)
        interpretation_member = self._new_interpretation_member(1)
        analyte = self._new_analyte(FIRST_ANALYTE_KEY)

        # Assume MSP spectra always have an Analyte, we can remove it later.
        # It is less common to assume that they will have Interpretation
        # or InterpretationMember component, so we will add these only if
        # we find attributes for them
        spectrum.add_analyte(analyte)
        spectrum.peak_list = peak_list

        #### Add special terms that we want to start off with
        for term in leader_terms:
            if term in attributes:
                spectrum.add_attribute(leader_terms[term], try_cast(attributes[term]))
                break
        else:
            logger.error(
                "Did not find any MSP leader terms (%s) for Spectrum Index = %s",
                leader_terms,
                spectrum.index,
            )

        #### Translate the rest of the known attributes and collect unknown ones
        attribute_rules = self._make_attribute_handlers()
        unknown_terms = attribute_rules.process_attributes(
            attributes, spectrum, analyte, interpretation, interpretation_member
        )

        #### Handle the uninterpretable terms
        for attribute in unknown_terms:
            self.unknown_attributes.add(attribute, attributes[attribute])
            if attributes[attribute] is None:
                spectrum.add_attribute(
                    CUSTOM_ATTRIBUTE_NAME, try_cast(attribute)
                )
            else:
                group_identifier = spectrum.get_next_group_identifier()
                spectrum.add_attribute(
                    CUSTOM_ATTRIBUTE_NAME,
                    try_cast(attribute),
                    group_identifier,
                )
                spectrum.add_attribute(
                    CUSTOM_ATTRIBUTE_VALUE,
                    try_cast(attributes[attribute]),
                    group_identifier,
                )

        if interpretation_member:
            interpretation.add_member_interpretation(interpretation_member)

        if not self._is_analyte_defined(analyte):
            self._hoist_analyte_attributes_on_rejection(analyte, spectrum)
            analyte.clear()
            spectrum.remove_analyte(analyte.id)
        elif analyte:
            spectrum.add_analyte(analyte)
            interpretation.add_analyte(analyte)
            spectrum.add_interpretation(interpretation)
        return spectrum

    def get_spectrum(
        self, spectrum_number: int = None, spectrum_name: str = None
    ) -> Spectrum:
        # keep the two branches separate for the possibility that this is not possible with all
        # index schemes.
        if spectrum_number is not None:
            if spectrum_name is not None:
                raise ValueError("Provide only one of spectrum_number or spectrum_name")
            index_record = self.index.record_for(spectrum_number)
            offset = index_record.offset
        elif spectrum_name is not None:
            index_record = self.index.record_for(spectrum_name)
            spectrum_number = index_record.number
            offset = index_record.offset
        else:
            raise ValueError(
                "Must provide either spectrum_number or spectrum_name argument"
            )
        buffer = self._get_lines_for(offset)
        spectrum = self._parse(buffer, index_record.index)
        return spectrum

    def summarize_parsing_errors(self) -> Dict:
        errors = super().summarize_parsing_errors()
        errors["unknown attributes"] = DefaultDict(dict)
        for k, v in self.unknown_attributes.items():
            errors["unknown attributes"][k] = v
        return errors


def _parse_fraction(x: str) -> float:
    a, b = x.split("/")
    return int(a) / int(b)


class MSPSpectralLibraryWriter(SpectralLibraryWriterBase):
    file_format = "msp"

    analyte_keys = {
        MOLECULAR_FORMULA: "Formula",
        "MS:1003044|number of missed cleavages": "MC",
        "MS:1001471|peptide modification details": "Mods",
        "MS:1003043|number of residues": "Naa",
        PRECURSOR_MZ: "PrecursorMonoisoMZ",
        "MS:1003054|theoretical average m/z": "Mz_av",
        PROFORMA_SEQ: "ProForma",
        STRIPPED_PEPTIDE_TERM: "Peptide",
        CHARGE_STATE: "Charge",
    }

    for species_name, keys in species_map.items():
        analyte_keys[tuple(keys[0])] = ("Organism", species_name)

    modification_map = {v: k for k, v in MODIFICATION_NAME_MAP.items()}

    spectrum_keys = {
        CHARGE_STATE: "Charge",
        (
            SPECTRUM_AGGREGATION_TYPE,
            SINGLETON_SPECTRUM,
        ): "Single",
        (
            SPECTRUM_AGGREGATION_TYPE,
            CONSENSUS_SPECTRUM,
        ): "Consensus",
        SCAN_NUMBER: "Scan",
        SOURCE_FILE: "Origfile",
        const.SAMPLE_NAME: "Sample",
        const.FILTER_STRING: "Filter",
        const.PRECURSOR_APEX_INTENSITY: "Precursor1MaxAb",
        const.ISOLATION_WINDOW_PRECURSOR_PURITY: "Purity",
        const.BASE_PEAK_INTENSITY: "BasePeak",
        "MS:1002599|splash key": "Splash",
        "MS:1003289|intensity of highest unassigned peak": "max_unassigned_ab",
        TOP_20_UNASSIGNED_INTENSITY_FRACTION: "Unassigned",
        TOTAL_UNASSIGNED_INTENSITY_FRACTION: "Unassign_all",
        NUM_UNASSIGNED_PEAKS_IN_TOP_20: "top_20_num_unassigned_peaks",
        (
            "MS:1000044|dissociation method",
            "MS:1000422|beam-type collision-induced dissociation",
        ): "HCD",
        (
            "MS:1000044|dissociation method",
            "MS:1002472|trap-type collision-induced dissociation",
        ): "CID",
    }

    # TODO: add these
    interpretation_keys = {
        const.Q_VALUE: "Q-value",
    }

    def __init__(self, filename, **kwargs):
        super(MSPSpectralLibraryWriter, self).__init__(filename)
        self._coerce_handle(self.filename)

    def write_header(self, library: SpectralLibraryBackendBase):
        pass

    def write_attribute_set(
        self, attribute_set: AttributeSet, attribute_set_type: AttributeSetTypes
    ):
        pass

    def _format_value(self, value):
        if isinstance(value, str):
            if not (value.startswith('"') and value.endswith('"')):
                value = f'"{value}"'
        return str(value)

    def _proforma_to_mods(self, proforma_seq: str) -> str:
        if isinstance(proforma_seq, proforma.ProForma):
            parsed = proforma_seq
        else:
            parsed = proforma.ProForma.parse(proforma_seq)
        mods = [(i, tok) for i, tok in enumerate(parsed) if tok[1]]
        if mods:
            tokens = []
            for i, mod_site in mods:
                tokens.append(str(i))
                tokens.append(mod_site[0])
                mod = mod_site[1][0]
                mod_name = self.modification_map.get(mod.name, mod.name)
                tokens.append(mod_name)
            return f'Mods={len(mods)}({",".join(tokens)})'
        else:
            return "Mods=0"

    def _protein_to_comments(self, analyte: Analyte) -> List[str]:
        acc = []
        protein: ProteinDescription
        for protein in analyte.proteins:
            accession = None
            pre = None
            post = None
            if protein.accession:
                accession = protein.accession
            if protein.flanking_n_terminal_residue:
                pre = protein.flanking_n_terminal_residue
            if protein.flanking_c_terminal_residue:
                post = protein.flanking_c_terminal_residue
            if accession:
                token = accession
                if token.startswith('"'):
                    token = token.strip('"')
                if pre or post:
                    token += f"(pre={pre or '-'},post={post or '-'})"
                acc.append(f"Protein={self._format_value(token)}")
                if protein.number_of_enzymatic_termini == 2:
                    if protein.cleavage_agent == "MS:1001251|Trypsin":
                        acc.append("Pep=Tryptic")
                elif protein.number_of_enzymatic_termini == 1:
                    if protein.cleavage_agent == "MS:1001251|Trypsin":
                        acc.append("Pep=SemiTryptic")
                if protein.missed_cleavages is not None:
                    acc.append(f"MC={self._format_value(protein.missed_cleavages)}")
                break
        return acc

    def _build_comments(
        self, spectrum: Spectrum, attribute_container: Attributed, rule_map: Dict
    ) -> List[Tuple[str, str]]:
        accumulator = []

        for attr_name, msp_name in rule_map.items():
            if isinstance(attr_name, tuple):
                if attribute_container.has_attribute(attr_name[0]):
                    value = attribute_container.get_attribute(attr_name[0])
                    if value == attr_name[1]:
                        if isinstance(msp_name, str):
                            accumulator.append(msp_name)
                        elif isinstance(msp_name, (list, tuple)) and len(msp_name) == 2:
                            accumulator.append("=".join(msp_name))
                        else:
                            raise TypeError(
                                f"Can't infer conversion for {msp_name} given {attr_name}"
                            )
            elif attribute_container.has_attribute(attr_name):
                value = attribute_container.get_attribute(attr_name)
                if isinstance(value, list):
                    logger.warn(
                        "Spectrum %r contains multiple values for %r, only the first will be saved",
                        spectrum.name,
                        attr_name,
                    )
                    accumulator.append(f"{msp_name}={self._format_value(value[0])}")
                else:
                    accumulator.append(f"{msp_name}={self._format_value(value)}")
        return accumulator

    def build_spectrum_comments(self, spectrum: Spectrum) -> List[Tuple[str, str]]:
        accumulator = self._build_comments(spectrum, spectrum, self.spectrum_keys)
        if spectrum.analytes:
            analyte = spectrum.get_analyte("1")
            accumulator += self._build_comments(spectrum, analyte, self.analyte_keys)
            if analyte.peptide:
                accumulator.append(self._proforma_to_mods(analyte.peptide))
            accumulator += self._protein_to_comments(analyte)
        if spectrum.interpretations:
            interp = spectrum.get_interpretation("1")
            accumulator += self._build_comments(
                spectrum, interp, self.interpretation_keys
            )
        return accumulator

    def write_spectrum(self, spectrum: Spectrum):
        if len(spectrum.analytes) > 1:
            logger.warning(
                "Spectrum %r contains multiple analytes, MSP will only contain the first",
                spectrum.name,
            )
        analyte = spectrum.get_analyte("1")
        self.handle.write(f"Name: {spectrum.name}\n")
        self.handle.write(f"MW: {analyte.mass}\n")
        self.handle.write(
            f"Comment: {' '.join(self.build_spectrum_comments(spectrum))}\n"
        )
        self._write_peaks(spectrum)
        self.handle.write("\n")

    def _format_annotation(self, annot: annotation.IonAnnotationBase):
        parts = []
        if isinstance(annot, annotation.PeptideFragmentIonAnnotation):
            parts.append(f"{annot.series}{annot.position}")
        elif isinstance(annot, annotation.ImmoniumIonAnnotation):
            parts.append(
                f"I{annot.amino_acid}{annot.modification if annot.modification else ''}"
            )
        elif isinstance(annot, annotation.ReporterIonAnnotation):
            parts.append(annot.reporter_label)
        if not parts:
            return "?"
        if annot.neutral_losses:
            f = annotation.NeutralName.combine(annot.neutral_losses)
            if f[0] not in ("-", "+"):
                f = "+" + f
            parts.append(f)
        if annot.adducts:
            parts.append(f"[{annotation.combine_formula(annot.adducts)}]")
        if annot.charge > 1:
            parts.append(f"^{annot.charge}")
        if annot.isotope:
            if annot.isotope > 0:
                parts.append(f"+{annot.isotope}i")
            else:
                parts.append(f"{annot.isotope}i")
        if annot.mass_error:
            parts.append(f"/{annot.mass_error}")
        return "".join(parts)

    def _write_peaks(self, spectrum: Spectrum):
        self.handle.write(f"Num peaks: {len(spectrum.peak_list)}\n")
        for peak in spectrum.peak_list:
            annot = self._format_annotation(peak[2][0]) if peak[2] else "?"
            self.handle.write(f'{peak[0]:0.4f}\t{peak[1]:0.4f}\t"{annot}"\n')

    def close(self):
        if not self.handle.closed:
            self.handle.close()



import os
import unittest
import tempfile

from mzspeclib.spectrum_library import SpectrumLibrary

from .common import datafile
from .test_library_backend import LibraryBehaviorBase


class TestSpectrumLibrary(unittest.TestCase, LibraryBehaviorBase):
    library_cls = SpectrumLibrary
    test_file = datafile("chinese_hamster_hcd_selected_head.mzlb.txt")

    def _open_library(self):
        lib = self.library_cls(filename=self.test_file)
        return lib



import logging
import re
from typing import Dict


class LevelAwareColoredLogFormatter(logging.Formatter):
    try:
        from colorama import Fore, Style
        # GREY = Fore.WHITE
        GREY = ''
        BLUE = Fore.BLUE
        GREEN = Fore.GREEN
        YELLOW = Fore.YELLOW
        RED = Fore.RED
        BRIGHT = Style.BRIGHT
        DIM = Style.DIM
        BOLD_RED = Fore.RED + Style.BRIGHT
        RESET = Style.RESET_ALL
    except ImportError:
        GREY = ''
        BLUE = ''
        GREEN = ''
        YELLOW = ''
        RED = ''
        BRIGHT = ''
        DIM = ''
        BOLD_RED = ''
        RESET = ''

    def _colorize_field(self, fmt: str, field: str, color: str) -> str:
        return re.sub("(" + field + ")", color + r"\1" + self.RESET, fmt)

    def _patch_fmt(self, fmt: str, level_color: str) -> str:
        fmt = self._colorize_field(fmt, r"%\(asctime\)s", self.GREEN)
        fmt = self._colorize_field(fmt, r"%\(name\).*?s", self.BLUE)
        fmt = self._colorize_field(fmt, r"%\(message\).*?s", self.GREY)
        if level_color:
            fmt = self._colorize_field(fmt, r"%\(levelname\).*?s", level_color)
        return fmt

    def __init__(self, fmt, level_color=None, **kwargs):
        fmt = self._patch_fmt(fmt, level_color=level_color)
        super().__init__(fmt, **kwargs)


class ColoringFormatter(logging.Formatter):
    level_to_color = {
        logging.INFO: LevelAwareColoredLogFormatter.GREEN,
        logging.DEBUG: LevelAwareColoredLogFormatter.GREY + LevelAwareColoredLogFormatter.DIM,
        logging.WARN: LevelAwareColoredLogFormatter.YELLOW + LevelAwareColoredLogFormatter.BRIGHT,
        logging.ERROR: LevelAwareColoredLogFormatter.BOLD_RED,
        logging.CRITICAL: LevelAwareColoredLogFormatter.BOLD_RED,
        logging.FATAL: LevelAwareColoredLogFormatter.RED + LevelAwareColoredLogFormatter.DIM,
    }

    _formatters: Dict[int, LevelAwareColoredLogFormatter]

    def __init__(self, fmt: str, **kwargs):
        self._formatters = {}
        for level, style in self.level_to_color.items():
            self._formatters[level] = LevelAwareColoredLogFormatter(fmt, level_color=style, **kwargs)

    def format(self, record: logging.LogRecord) -> str:
        fmtr = self._formatters[record.levelno]
        return fmtr.format(record)



import warnings
import logging

from typing import Any, Dict, Iterator, Optional, List, DefaultDict, Union

from numbers import Integral
from collections import defaultdict

from mzspeclib.index.base import IndexRecordBase

from .base import IndexBase, IndexRecordBase, IndexInitializedRecord

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class _IndexAttr:
    __slots__ = ()

    def get(self, key: str, default=None) -> Any:
        if self.attributes is not None:
            return self.attributes.get(key, default)
        return default

    def set(self, key: str, value: Any):
        if self.attributes is not None:
            self.attributes[key] = value
        else:
            self.attributes = {key: value}


class IndexRecord(IndexRecordBase, _IndexAttr):
    """
    A spectrum index record.

    Attributes
    ----------
    number : int
        A numerical identifier for the spectrum
    offset : int
        The offset in the file to reach the spectrum (in bytes if appropriate)
    name : str,
        A text identifier for this spectrum.
    analyte : str, optional
        A text representation of the analyte for that record
    attributes : Dict[str, Any], optional
        A key-value pair collection of this record.
    """

    __slots__ = ('number', 'offset', 'name', 'analyte', 'index', 'attributes')

    number: int
    offset: int
    name: str
    index: int
    analyte: Any
    attributes: Optional[Dict[str, Any]]

    def __init__(self, number, offset, name, analyte, index: int=None, attributes=None):
        self.number = number
        self.offset = offset
        self.name = name
        self.analyte = analyte
        self.index = index
        self.attributes = attributes

    def __repr__(self):
        template = f"{self.__class__.__name__}({self.number}, {self.offset}, {self.name}, {self.analyte}, {self.index}, {self.attributes})"
        return template

    def __eq__(self, other):
        if self.number != other.number:
            return False
        elif self.offset != other.offset:
            return False
        elif self.name != other.name:
            return False
        elif self.analyte != other.analyte:
            return False
        if bool(self.attributes) == bool(other.attributes):
            if bool(self.attributes) and self.attributes != other.attributes:
                return False
            # Implicitly allow None and empty dictionaries to be the same
        return True

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.name)

    def to_dict(self) -> Dict:
        return {
            k: getattr(self, k, None) for k in self.__slots__
        }

    @classmethod
    def from_dict(cls, state: Dict) -> 'IndexRecord':
        return cls(**state)


class ClusterIndexRecord(IndexRecordBase, _IndexAttr):
    """
    A spectrum cluster index record.

    Attributes
    ----------
    number : int
        A numerical identifier for this spectrum.
    offset : int
        The offset in the file to reach the spectrum (in bytes if appropriate)
    attributes : Dict[str, Any], optional
        A key-value pair collection of this record
    """

    __slots__ = ('number', 'offset', 'attributes')

    def __init__(self, number, offset, attributes=None):
        self.number = number
        self.offset = offset
        self.attributes = attributes

    def __repr__(self):
        template = f"{self.__class__.__name__}({self.number}, {self.offset}, {self.attributes})"
        return template

    def __eq__(self, other):
        if self.number != other.number:
            return False
        elif self.offset != other.offset:
            return False
        if bool(self.attributes) == bool(other.attributes):
            if bool(self.attributes) and self.attributes != other.attributes:
                return False
            # Implicitly allow None and empty dictionaries to be the same
        return True

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.name)

    def to_dict(self) -> Dict:
        return {
            k: getattr(self, k, None) for k in self.__slots__
        }

    @classmethod
    def from_dict(cls, state: Dict) -> 'ClusterIndexRecord':
        return cls(**state)


class MemoryIndex(IndexBase):
    """
    An in-memory data structure for holding spectrum metadata and offsets.

    Attributes
    ----------
    records : List[:class:`IndexRecord`]
        The index entries for spectra
    cluster_records : List[:class:`ClusterIndexRecord`]
        The index entries for clusters
    metadata : Dict[str, Any]
        Arbitrary metadata about the library or the index
    """

    records: List[IndexRecord]
    cluster_records: List[ClusterIndexRecord]
    metadata: Dict[str, Any]

    _dirty: bool
    _by_key: Dict[int, IndexRecord]
    _by_key_cluster: Dict[int, ClusterIndexRecord]
    _by_name: DefaultDict[str, List[IndexRecord]]
    _by_attr: DefaultDict[str, DefaultDict[Any, List[IndexRecord]]]

    @classmethod
    def from_filename(cls, filename, library=None):
        inst = cls()
        return IndexInitializedRecord(inst, False)

    def __init__(self, records=None, cluster_records=None, metadata=None):
        self.records = list(records or [])
        self.cluster_records = list(cluster_records or [])
        self._by_name = defaultdict(list)
        self._by_key = {}
        self._by_key_cluster = {}
        self._by_attr = defaultdict(lambda: defaultdict(list))
        self.metadata = metadata or {}
        self._dirty = True

    def iter_clusters(self) -> Iterator[IndexRecordBase]:
        """Iterate over cluster entries in the index."""
        return iter(self.cluster_records)

    def iter_spectra(self):
        """Iterate over spectrum entries in the index."""
        return iter(self.records)

    def __len__(self):
        return len(self.records)

    def search(self, i=None, **kwargs):
        if self._dirty:
            self._update_index()
        if i is None and kwargs:
            # Executing attribute query
            raise NotImplementedError()
        if isinstance(i, Integral):
            try:
                return self._by_key[i]
            except IndexError as err:
                raise KeyError(i) from err
        elif isinstance(i, slice):
            start = i.start
            stop = i.stop
            if start is None:
                start = min(self._by_key) if self._by_key else 0
            if stop is None:
                stop = max(self._by_key) if self._by_key else 0
            return [self._by_key[i] for i in range(start, stop) if i in self._by_key]
        if i in self._by_name:
            records = self._by_name[i]
            if len(records) == 1:
                return records[0]
            else:
                return records
        else:
            raise KeyError(i)

    def search_clusters(self, i=None, **kwargs):
        if self._dirty:
            self._update_index()

        if i is None and kwargs:
            # Executing attribute query
            raise NotImplementedError()

        if isinstance(i, Integral):
            return self._by_key_cluster[i]

        elif isinstance(i, slice):
            start = i.start
            stop = i.stop
            if start is None:
                start = min(self._by_key_cluster) if self._by_key_cluster else 0
            if stop is None:
                stop = max(self._by_key_cluster) if self._by_key_cluster else 0
            return [self._by_key_cluster[i] for i in range(start, stop) if i in self._by_key_cluster]

    def __getitem__(self, i):
        return self._get_by_index(i)

    def _get_by_index(self, i: Union[int, slice]) -> Union[IndexRecord, List[IndexRecord]]:
        return self.records[i]

    def _update_index(self):
        self.records.sort(key=lambda x: x.number)

        self._by_key.clear()
        self._by_name = defaultdict(list)
        for record in self:
            self._by_key[record.number] = record
            self._by_name[record.name].append(record)

        self.cluster_records.sort(key=lambda x: x.number)
        self._by_key_cluster.clear()
        for record in self.cluster_records:
            self._by_key_cluster[record.number] = record

        self._dirty = False

    def add(self, number: int, offset: int, name: str, analyte: Any, attributes=None):
        """
        Add a new entry to the spectrum index.

        Parameters
        ----------
        number : int
            A numerical identifier for this spectrum.
        offset : int
            The offset in the file to reach the spectrum (in bytes if appropriate)
        name : str,
            A text identifier for this spectrum.
        analyte : str, optional
            A text representation of the analyte for that record
        attributes : Dict[str, Any], optional
            A key-value pair collection of this record, currently not supported.
        """
        n = len(self.records)
        record = IndexRecord(number, offset, name, analyte, n, attributes)
        self.records.append(record)
        self._dirty = True

    def add_cluster(self, number: int, offset: int, attributes=None):
        """
        Add a new entry to the spectrum index.

        Parameters
        ----------
        number : int
            A numerical identifier for this spectrum.
        offset : int
            The offset in the file to reach the spectrum (in bytes if appropriate)
        attributes : Dict[str, Any], optional
            A key-value pair collection of this record, currently not supported.
        """
        record = ClusterIndexRecord(number, offset, attributes)
        self.cluster_records.append(record)
        self._dirty = True


    def commit(self):
        self._update_index()



"""A set of base types, functions and constants for implementing spectral library formats."""
import io
import csv
import enum
import logging
import os
import warnings

from typing import Any, Callable, Dict, Iterable, Optional, Union, List, Type, Iterator, TYPE_CHECKING
from pathlib import Path


from psims.controlled_vocabulary import Entity
from psims.controlled_vocabulary.controlled_vocabulary import (
    load_uo, load_unimod, load_psims)
from mzspeclib.cluster import SpectrumCluster

from mzspeclib.index import MemoryIndex, SQLIndex, IndexBase
from mzspeclib.spectrum import Spectrum
from mzspeclib.analyte import Analyte, Interpretation, InterpretationMember
from mzspeclib.attributes import Attributed, AttributedEntity, AttributeSet, AttributeManagedProperty
from mzspeclib.ontology import _VocabularyResolverMixin
from mzspeclib.const import (
    FORMAT_VERSION,
    LIBRARY_NAME,
    LIBRARY_IDENTIFIER,
    LIBRARY_VERSION,
    LIBRARY_URI,
    LIBRARY_DESCRIPTION,
    ANALYTE_MIXTURE,
    LIBRARY_SPECTRUM_INDEX,
    LIBRARY_SPECTRUM_KEY
)

from .utils import open_stream, _LineBuffer

if TYPE_CHECKING:
    from mzspeclib import SpectrumLibrary

logger = logging.getLogger(__name__.rsplit(".", 1)[0])
logger.addHandler(logging.NullHandler())

ANALYTE_MIXTURE_CURIE = ANALYTE_MIXTURE.split("|")[0]

DEFAULT_VERSION = '1.0'


class AttributeSetTypes(enum.Enum):
    """Attribute set type tags used as keys and section constants"""

    spectrum = enum.auto()
    analyte = enum.auto()
    interpretation = enum.auto()
    cluster = enum.auto()


class FormatInferenceFailure(ValueError):
    """Indicates that we failed to infer the format type for a spectral library."""


class SubclassRegisteringMetaclass(type):
    """
    A metaclass that registers instances instance types on the following attributes:
        - :attr:`file_format`
        - :attr:`format_name`

    It will also create :attr:`_file_extension_to_implementation` on the instance type
    if it does not already exist in the type's inheritance hierarchy.

    This is meant to be used to automate registration of new backends so as soon as the
    type is imported, it is registered.
    """

    def __new__(mcs, name, parents, attrs):
        new_type = type.__new__(mcs, name, parents, attrs)
        if not hasattr(new_type, "_file_extension_to_implementation"):
            new_type._file_extension_to_implementation = dict()

        file_extension = attrs.get("file_format")
        if file_extension is not None:
            if isinstance(file_extension, list):
                for ext in file_extension:
                    new_type._file_extension_to_implementation[ext] = new_type
            else:
                new_type._file_extension_to_implementation[file_extension] = new_type

        format_name = attrs.get("format_name")
        if format_name is not None:
            new_type._file_extension_to_implementation[format_name] = new_type
        else:
            attrs['format_name'] = file_extension
        return new_type

    def type_for_format(cls, format_or_extension: str) -> Type:
        """Get the implementing type for a specific format or file extension, if it exists"""
        return cls._file_extension_to_implementation.get(format_or_extension)


class _LibraryViewMixin:

    name = AttributeManagedProperty[str](LIBRARY_NAME)
    identifier = AttributeManagedProperty[str](LIBRARY_IDENTIFIER)
    description = AttributeManagedProperty[str](LIBRARY_DESCRIPTION)
    uri = AttributeManagedProperty[str](LIBRARY_URI)
    library_version = AttributeManagedProperty[str](LIBRARY_VERSION)

    @property
    def format_version(self):
        try:
            value = self.get_attribute(FORMAT_VERSION)
            return value
        except KeyError:
            value = DEFAULT_VERSION
            self.add_attribute(FORMAT_VERSION, value)
            return value


class SpectralLibraryBackendBase(AttributedEntity, _VocabularyResolverMixin, _LibraryViewMixin, metaclass=SubclassRegisteringMetaclass):
    """A base class for all spectral library formats."""

    file_format = None

    _file_extension_to_implementation: Dict[str,
                                            Type['SpectralLibraryBackendBase']] = {}
    _format_name_to_implementation: Dict[str,
                                         Type['SpectralLibraryBackendBase']] = {}

    index: IndexBase

    spectrum_attribute_sets: Dict[str, AttributeSet]
    analyte_attribute_sets: Dict[str, AttributeSet]
    interpretation_attribute_sets: Dict[str, AttributeSet]
    cluster_attribute_sets: Dict[str, AttributeSet]

    @classmethod
    def guess_from_filename(cls, filename: Union[str, Path, io.FileIO]) -> bool:
        """
        Guess if the file is of this type by inspecting the file's name and extension.

        Parameters
        ----------
        filename : str
            The path to the file to inspect.

        Returns
        -------
        bool:
            Whether this is an appropriate backend for that file.
        """
        if hasattr(filename, "name"):
            filename = filename.name
        if not isinstance(filename, (str, Path)):
            return False
        if filename.endswith(".gz"):
            filename = filename[:-3]
        if isinstance(cls.file_format, list):
            return any(filename.endswith(ext) for ext in cls.file_format)
        return filename.endswith(cls.file_format)

    @classmethod
    def guess_from_header(cls, filename: Union[str, Path, io.FileIO]) -> bool:
        """
        Guess if the file is of this type by inspecting the file's header section

        Parameters
        ----------
        filename : str
            The path to the file to open.

        Returns
        -------
        bool:
            Whether this is an appropriate backend for that file.
        """
        return False

    @classmethod
    def guess_implementation(cls, filename: Union[str, Path, io.FileIO], index_type=None,
                             **kwargs) -> 'SpectralLibraryBackendBase':
        """
        Guess the backend implementation to use with this file format.

        Parameters
        ----------
        filename : str
            The path to the spectral library file to open.
        index_type : type, optional
            The :class:`~.IndexBase` derived type to use for this file. If
            :const:`None` is provided, the instance will decide based upon
            :meth:`has_index_preference`.
        **kwargs
            Passed to implementation

        Returns
        -------
        SpectralLibraryBackendBase
        """
        for key, impl in cls._file_extension_to_implementation.items():
            try:
                if impl.guess_from_filename(filename):
                    return impl(filename, index_type=index_type, **kwargs)
            except TypeError:
                pass
            try:
                if impl.guess_from_header(filename):
                    return impl(filename, index_type=index_type, **kwargs)
            except (TypeError, UnicodeDecodeError):
                pass
        raise FormatInferenceFailure(f"Could not guess backend implementation for {filename}")

    def __init__(self, filename: Union[str, Path, io.FileIO]):
        self.filename = filename
        self.index = MemoryIndex()

        self.spectrum_attribute_sets = {
            "all": AttributeSet("all", [])
        }
        self.analyte_attribute_sets = {
            "all": AttributeSet("all", [])
        }
        self.interpretation_attribute_sets = {
            "all": AttributeSet("all", [])
        }
        self.cluster_attribute_sets = {
            "all": AttributeSet("all", [])
        }

        super().__init__(None)

    def _infer_lib_name(self) -> Optional[str]:
        if hasattr(self.filename, "name"):
            name = self.filename.name.replace(".gz", "").rsplit(".", 1)[0].split(os.sep)[-1]
        elif isinstance(self.filename, str):
            name = self.filename.replace(".gz", "").rsplit(".", 1)[0].split(os.sep)[-1]
        elif isinstance(self.filename, bytes):
            name = self.filename.decode("utf8").replace(".gz", "").rsplit(".", 1)[0].split(os.sep)[-1]
        else:
            name = None
        return name

    def read_header(self) -> bool:
        """
        Read just the header of the whole library

        Returns
        -------
        bool
        """
        raise NotImplementedError()

    def _new_spectrum(self) -> Spectrum:
        spec = Spectrum()
        attr_set = self.spectrum_attribute_sets.get("all")
        if attr_set:
            attr_set.apply(spec)
        return spec

    def _new_interpretation(self, id=None) -> Interpretation:
        interp = Interpretation(id)
        attr_set = self.interpretation_attribute_sets.get('all')
        if attr_set:
            attr_set.apply(interp)
        return interp

    def _new_interpretation_member(self, id=None) -> InterpretationMember:
        return InterpretationMember(id)

    def _new_analyte(self, id=None) -> Analyte:
        analyte = Analyte(id)
        attr_set = self.analyte_attribute_sets.get('all')
        if attr_set:
            attr_set.apply(analyte)
        return analyte

    def _new_cluster(self) -> SpectrumCluster:
        cluster = SpectrumCluster()
        attr_set = self.cluster_attribute_sets.get('all')
        if attr_set:
            attr_set.apply(cluster)
        return cluster

    def _analyte_interpretation_link(self, spectrum: Spectrum,
                                     interpretation: Interpretation):
        if (interpretation.has_attribute(ANALYTE_MIXTURE) and
            not interpretation.analytes):
            analyte_ids = interpretation.get_attribute(ANALYTE_MIXTURE)
            if isinstance(analyte_ids, str):
                term = self.find_term_for(ANALYTE_MIXTURE_CURIE)
                analyte_ids = term.value_type(analyte_ids)

            # TODO: Enforce this attribute is a string at the CV level
            # if isinstance(analyte_ids_term, int):
            #     analyte_ids = [analyte_ids_term]
            #     interpretation.replace_attribute(ANALYTE_MIXTURE_TERM, str(analyte_ids_term))
            # else:
            #     analyte_ids = analyte_ids_term.split(',')
            for analyte_id in analyte_ids:
                interpretation.add_analyte(spectrum.get_analyte(analyte_id))
        return interpretation

    def _default_interpretation_to_analytes(self, spectrum: Spectrum):
        for interpretation in spectrum.interpretations.values():
            if not interpretation.analytes:
                for analyte in spectrum.analytes.values():
                    interpretation.add_analyte(analyte)

    def _is_analyte_defined(self, analyte: Analyte) -> bool:
        """
        Check if the :class:`~.Analyte` has been sufficiently described
        by the attributes we've seen so far, either for a peptide or some
        other kind of molecule.

        If not, we'll do other work in :meth:`_hoist_analyte_attributes_on_rejection`
        to move information out of the :class:`~.Analyte` up to the :class:`~.Spectrum`
        and omit the analyte entirely.
        """
        peptide_attr_root = self.find_term_for("MS:1003050")
        molecular_attr_root = self.find_term_for("MS:1003033")
        for attrib in analyte:
            acc, _name = attrib.key.split("|")
            term = self.find_term_for(acc)
            if term.is_of_type(peptide_attr_root):
                return True
            if term.is_of_type(molecular_attr_root):
                return True
        return False

    def _hoist_analyte_attributes_on_rejection(self, analyte: Analyte, spectrum: Spectrum):
        """
        When an :class:`~.Analyte` is not defined well enough, move any terms
        that are ion-related up to the :class:`~.Spectrum`.
        """
        ion_selection_root = self.find_term_for("MS:1000455")
        attribute_groups_to_hoist = {}
        for attrib in analyte:
            acc, _name = attrib.key.split("|")
            term = self.find_term_for(acc)
            if attrib.group_id in attribute_groups_to_hoist:
                attribute_groups_to_hoist[attrib.group_id].append(attrib)
            if term.is_of_type(ion_selection_root):
                if attrib.group_id is None:
                    spectrum.add_attribute(attrib.key, attrib.value)
                else:
                    attribute_groups_to_hoist[attrib.group_id] = [attrib]

        for group in attribute_groups_to_hoist.values():
            spectrum.add_attribute_group(group)

    def get_spectrum(self, spectrum_number: int=None,
                     spectrum_name: str=None) -> Spectrum:
        """
        Retrieve a single spectrum from the library.

        Parameters
        ----------
        spectrum_number : int, optional
            The index of the spectrum in the library
        spectrum_name : str, optional
            The name of the spectrum in the library

        Returns
        -------
        :class:`~.Spectrum`
        """
        raise NotImplementedError()

    def get_cluster(self, cluster_number: int) -> SpectrumCluster:
        """
        Retrieve a single spectrum cluster from the library.

        Parameters
        ----------
        cluster_number : int, optional
            The index of the cluster in the library

        Returns
        -------
        :class:`~.SpectrumCluster`
        """
        raise NotImplementedError()

    def find_spectra(self, specification, **query_keys):
        raise NotImplementedError()

    def create_index(self) -> int:
        """
        Populate the spectrum index.

        This method may produce a large amount of file I/O.

        Returns
        -------
        n_spectra: int
            The number of entries read
        """
        raise NotImplementedError()

    def __iter__(self) -> Iterator[Spectrum]:
        if self.index:
            for record in self.index:
                yield self.get_spectrum(record.number)
        else:
            return self.read()

    def __len__(self):
        return len(self.index)

    def __getitem__(self, i) -> Union[Spectrum, List[Spectrum]]:
        record = self.index[i]
        if isinstance(record, list):
            result = [self.get_spectrum(rec.number) for rec in record]
        else:
            result = self.get_spectrum(record.number)
        return result

    @classmethod
    def has_index_preference(cls, filename: Union[str, Path, io.FileIO]) -> Type[IndexBase]:
        """
        Check if this backend prefers a particular index for this file.

        The base implementation checks to see if there is a SQL index
        for the filename provided, and if so, prefers :class:`~.SQLIndex`.
        Otherwise, prefers :class:`~.MemoryIndex`.

        Parameters
        ----------
        filename: str
            The name of the file to open.

        Returns
        -------
        index_type: type
            Returns a :class:`~.IndexBase` derived type which this backend
            would prefer to use.
        """
        try:
            if SQLIndex.exists(filename):
                return SQLIndex
            return MemoryIndex
        except Exception:
            return MemoryIndex

    def read(self) -> Iterator[Union[Spectrum, SpectrumCluster]]:
        """
        Create an sequential iterator over the spectrum library.

        Yields
        ------
        entry :  Union[:class:`~.Spectrum`, :class:`~.SpectrumCluster`]
        """
        raise NotImplementedError()

    def _add_attribute_set(self, attribute_set: AttributeSet,
                           attribute_set_type: AttributeSetTypes):
        if attribute_set_type == AttributeSetTypes.spectrum:
            self.spectrum_attribute_sets[attribute_set.name] = attribute_set
        elif attribute_set_type == AttributeSetTypes.analyte:
            self.analyte_attribute_sets[attribute_set.name] = attribute_set
        elif attribute_set_type == AttributeSetTypes.interpretation:
            self.interpretation_attribute_sets[attribute_set.name] = attribute_set
        elif attribute_set_type == AttributeSetTypes.cluster:
            self.cluster_attribute_sets[attribute_set.name] = attribute_set
        else:
            raise ValueError(f"Could not map {attribute_set_type}")

    def summarize_parsing_errors(self) -> Dict:
        """Retrieve a free-form description of parsing errors"""
        return {}


guess_implementation = SpectralLibraryBackendBase.guess_implementation


class _PlainTextSpectralLibraryBackendBase(SpectralLibraryBackendBase):

    def __init__(self, filename: Union[str, Path, io.FileIO], index_type=None,
                 read_metadata: bool=True, create_index: bool=True):
        if index_type is None and create_index:
            index_type = self.has_index_preference(filename)

        super(_PlainTextSpectralLibraryBackendBase, self).__init__(filename)

        if index_type is not None:
            self.index, was_initialized = index_type.from_filename(filename)
            if not was_initialized and create_index:
                self.create_index()
        if read_metadata:
            self.read_header()

    def _coerce_handle(self, filename_or_stream):
        if hasattr(filename_or_stream, 'read'):
            self.handle = filename_or_stream
        else:
            self.handle = open_stream(filename_or_stream, 'rt')

    def _buffer_from_stream(self, stream: io.IOBase) -> List:
        """
        Collect data from the readable stream until
        a complete spectrum entry has been observed.

        Parameters
        ----------
        stream: file-like
            Theinput file stream to read from.

        Returns
        -------
        line_buffer: List[str]
            A list of lines read from the input stream.
        """
        raise NotImplementedError()

    def read(self) -> Iterator[Spectrum]:
        with open_stream(self.filename, 'rb') as stream:
            i = 0
            match, offset = self._parse_header_from_stream(stream)
            if not match:
                raise ValueError("Could not locate valid header")
            else:
                stream.seek(offset)
            buffering_stream = _LineBuffer(stream, encoding="utf8")
            while True:
                # Will clip the first line of the next spectrum. Needs work
                buffer = self._buffer_from_stream(buffering_stream)

                # If the buffer is only a single line, then we must have reached
                # the end, so we're done. We're done because the buffering stream
                # will contain exactly one line (the buffered line)
                if len(buffer) <= 1:
                    break
                buffering_stream.push_line()
                yield self._parse(buffer, i)
                i += 1

    def _get_lines_for(self, offset: int) -> List[str]:
        filename = self.filename
        is_file_like_object = isinstance(filename, io.IOBase)

        infile = open_stream(filename, 'r')

        infile.seek(offset)
        spectrum_buffer = self._buffer_from_stream(infile)
        #### We will end up here if this is the last spectrum in the file
        if not is_file_like_object:
            infile.close()
        else:
            infile.detach()
        return spectrum_buffer

    def _parse(self, buffer: Iterable, spectrum_index: int=None):
        raise NotImplementedError()

    def search(self, specification, **query_keys) -> List[Spectrum]:
        records = self.index.search(specification, **query_keys)
        if not isinstance(records, list):
            records = [records]
        spectra = []
        for record in records:
            buffer = self._get_lines_for(record.offset)
            spectrum = self._parse(buffer, record.number)
            spectra.append(spectrum)
        return spectra


class _CSVSpectralLibraryBackendBase(SpectralLibraryBackendBase):
    _delimiter: str
    _header: List[str]

    _required_columns: List[str] = None

    @classmethod
    def guess_from_header(cls, filename) -> bool:
        with open_stream(filename, 'rt') as fh:
            line = fh.readline()
            tokens = line.split('\t')
            if len(tokens) > 3:
                warnings.warn(
                    "This file looks like a TSV file, but it's not possible to say definitively what"
                    " type from this alone. Please explicitly specify the format.")
                return False
        return False

    def __init__(self, filename: Union[str, Path, io.FileIO], index_type=None, delimiter='\t',
                 read_metadata: bool=True, create_index: bool = True, ** kwargs):
        if index_type is None:
            index_type = self.has_index_preference(filename)
        self._delimiter = delimiter
        self._headers = None
        super().__init__(filename)
        self.filename = filename

        self.read_header()
        self.index, was_initialized = index_type.from_filename(filename)
        if not was_initialized and create_index:
            self.create_index()

    def read_header(self) -> bool:
        self._read_header_line()
        return True

    def _read_header_line(self):
        headers = None
        with open_stream(self.filename) as stream:
            reader = csv.reader(stream, delimiter=self._delimiter)
            headers = next(reader)
            stream.seek(0)
        self._headers = headers
        if headers and self._required_columns:
            missing_required = set(self._required_columns) - set(headers)
            if missing_required:
                raise TypeError(
                    f"{self.format_name} requires column{'s' if len(missing_required) > 1 else ''} "
                    f"{', '.join(missing_required)}, but were not found."
                )

    def _open_reader(self, stream: io.TextIOBase) -> Union[Iterator[Dict[str, Any]], csv.DictReader]:
        return csv.DictReader(stream, fieldnames=self._headers, delimiter='\t')

    def get_spectrum(self, spectrum_number: int = None, spectrum_name: str = None) -> Spectrum:
        # keep the two branches separate for the possibility that this is not possible with all
        # index schemes.
        if spectrum_number is not None:
            if spectrum_name is not None:
                raise ValueError("Provide only one of spectrum_number or spectrum_name")
            offset = self.index.offset_for(spectrum_number)
        elif spectrum_name is not None:
            offset = self.index.offset_for(spectrum_name)
        else:
            raise ValueError("Must provide either spectrum_number or spectrum_name argument")
        buffer = self._get_lines_for(offset)
        spectrum = self._parse_from_buffer(buffer, spectrum_number)
        return spectrum

    def _batch_rows(self, iterator: Iterator[Dict[str, Any]]) -> Iterator[List[Dict[str, Any]]]:
        """
        Gather successive rows by some shared key into a batch to be parsed into a single
        :class:`~.Spectrum` instance.

        This assumes there are no interleaved rows across spectra.

        Parameters
        ----------
        iterator : Iterator[Dict[str, Any]]
            An iterator over rows of the underlying CSV as dictionaries

        Yields
        ------
        batch : List[Dict[str, Any]]
            The next batch of rows corresponding to a single spectrum
        """
        raise NotImplementedError()

    def _get_lines_for(self, offset: int) -> List[Dict[str, Any]]:
        with open_stream(self.filename, 'r') as infile:
            infile.seek(offset)
            reader = self._open_reader(infile)
            spectrum_buffer = next(self._batch_rows(reader))
            #### We will end up here if this is the last spectrum in the file
        return spectrum_buffer

    def read(self) -> Iterator[Spectrum]:
        with open_stream(self.filename, 'rt') as stream:
            i = 0
            reader = self._open_reader(stream)
            if self._headers:
                # Skip the header line if we've already parsed them
                _ = next(reader)
            buffering_reader = self._batch_rows(reader)
            for i, buffer in enumerate(buffering_reader):
                yield self._parse_from_buffer(buffer, i)


class SpectralLibraryWriterBase(_VocabularyResolverMixin, metaclass=SubclassRegisteringMetaclass):
    """
    A base type for spectral library writers.

    This type implements the context manager protocol, controlling the closing of the
    enclosed IO stream.

    Attributes
    ----------
    filename : str, :class:`pathlib.Path`, or :class:`io.IOBase`
    """

    _already_started_writing: bool = False

    def __init__(self, filename, **kwargs):
        self.filename = filename
        super().__init__(**kwargs)

    def _filter_attributes(self, attributes: Attributed,
                           filter_fn: Callable) -> Iterable:
        if isinstance(attributes, AttributedEntity):
            attributes = attributes.attributes
        for attrib in attributes:
            if filter_fn(attrib):
                yield attrib

    def _not_analyte_mixture_term(self, attrib):
        if attrib:
            key = attrib[0]
            if key == ANALYTE_MIXTURE:
                return False
        return True

    def _not_entry_index(self, attrib):
        if attrib:
            key = attrib[0]
            if key == LIBRARY_SPECTRUM_INDEX:
                return False
        return True

    def _not_entry_key_or_index(self, attrib):
        if attrib:
            key = attrib[0]
            if key in (LIBRARY_SPECTRUM_INDEX, LIBRARY_SPECTRUM_KEY):
                return False
        return True

    def _coerce_handle(self, filename_or_stream):
        if hasattr(filename_or_stream, 'write'):
            self.handle = filename_or_stream
        else:
            self.handle = open(filename_or_stream, 'wt')

    def write_library(self, library: Union[SpectralLibraryBackendBase, "SpectrumLibrary"]):
        """
        Write out the entire library.

        Parameters
        ----------
        library : :class:`SpectralLibraryBackendBase` or :class:`SpectrumLibrary`
            The library to write out.

        Raises
        ------
        :class:`ValueError` :
            If the writer has already started writing one library, an error will be
            raised.
        """
        if self._already_started_writing:
            raise ValueError("Cannot write a new library, already started writing")
        self._already_started_writing = True

        self.write_header(library)
        n = len(library)
        step = max(min(n // 100, 5000), 1)
        ident = ''
        i = 0
        for i, entry in enumerate(library.read()):
            if i % step == 0 and i:
                if isinstance(entry, SpectrumCluster):
                    tag = "cluster "
                else:
                    tag = ""
                try:
                    ident = f"{tag}{entry.key}:{entry.name}"
                except Exception:
                    ident = f"{tag}{entry.key}"
                logger.info(f"Wrote {ident} {i}/{n} ({i / n * 100.0:0.2f}%)")
            if isinstance(entry, Spectrum):
                self.write_spectrum(entry)
            elif isinstance(entry, SpectrumCluster):
                self.write_cluster(entry)
            else:
                raise TypeError(f"Don't know how to save {entry.__class__}")

        i = n
        logger.info(f"Wrote {n} spectra")

    def write_spectrum(self, spectrum: Spectrum):
        """
        Write out a :class:`~.Spectrum` and all of its
        components.

        Parameters
        ----------
        spectrum : :class:`~.Spectrum`
            The spectrum to write.
        """
        raise NotImplementedError()

    def write_cluster(self, cluster: SpectrumCluster):
        """
        Write out a :class:`~.SpectrumCluster` and all of its
        components.

        Parameters
        ----------
        cluster : :class:`~.SpectrumCluster`
            The spectrum cluster to write.
        """
        raise NotImplementedError()

    def __enter__(self) -> 'SpectralLibraryWriterBase':
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __del__(self):
        self.close()

    def close(self):
        """
        Close the library writer, performing any necessary finalization.

        This is called automatically when :meth:`__exit__` is called.
        """
        pass


class LibraryIterator(AttributedEntity, _LibraryViewMixin, Iterator[Spectrum]):
    """An iterator wrapper for a library source that doesn't permit random access"""

    backend: SpectralLibraryBackendBase
    attributes: Attributed
    iter: Iterator[Spectrum]
    _buffer: Optional[Spectrum]

    def __init__(self, backend: SpectralLibraryBackendBase) -> None:
        self.backend = backend
        self.attributes = backend
        self.iter = backend.read()
        try:
            self._buffer = next(self.iter)
        except StopIteration:
            self._buffer = None

    def __iter__(self):
        return self

    def __next__(self) -> Spectrum:
        if self._buffer is not None:
            result = self._buffer
            self._buffer = None
            return result
        return next(self.iter)



import argparse
import itertools
import json

import csv
import xlsxwriter

import psims

cv = psims.load_psims()

def rule_to_table_rows(rule: dict):
    row_base = {
        "id": rule["id"],
        "path": rule["path"],
        "level": rule.get("level", rule.get("requirement_level")),
        "notes": rule.get("notes"),
        "combination_logic": rule.get("combination_logic"),
    }

    if "condition" in rule:
        condition = rule["condition"]
        row_base["condition_name"] = condition["name"]
        row_base["condition_accession"] = condition["accession"]
        row_base["condition_allow_children"] = condition["allow_children"]
        if "value" in condition:
            cond_value = condition["value"]
            for k, v in cond_value.items():
                row_base[f"condition_value_{k}"] = k
                row_base[f"condition_value_{v}"] = v

    if "attr" not in rule:
        yield row_base
    else:
        for attr in rule["attr"]:
            row = row_base.copy()
            row["attribute_name"] = attr["name"]
            row["attribute_accession"] = attr["accession"]
            row["allow_children"] = attr.get("allow_children")
            row['repeatable'] = attr.get('repeatable', False)
            term = cv[attr["accession"]]
            row["description"] = term.definition
            units = ";".join([f"{rel.accession}|{rel.comment}" for rel in term.get("has_units", [])])
            if units:
                row["has_units"] = units

            row['default_unit'] = attr.get("default_unit")
            row["attr_notes"] = attr.get("notes")
            if "value" in attr:
                value_rule = attr["value"]
                if isinstance(value_rule, str):
                    row["value_rule"] = value_rule
                else:
                    row["value_rule"] = value_rule["name"]
                    value_constraint = None
                    for k, v in value_rule.items():
                        if k != "name":
                            if value_constraint is not None:
                                raise ValueError("Too many properties in value rule")
                            else:
                                value_constraint = v
                    row["value_constraint"] = value_constraint
            if "condition" in attr:
                condition = attr["condition"]
                row["condition_name"] = condition["name"]
                row["condition_accession"] = condition["accession"]
                row["condition_allow_children"] = condition["allow_children"]
                if "value" in condition:
                    cond_value = condition["value"]
                    for k, v in cond_value.items():
                        row[f"condition_value_{k}"] = k
                        row[f"condition_value_{v}"] = v

            yield row


def infer_headers(rows):
    seen = set(("notes",))
    keys = []
    for row in rows:
        for k in row:
            if k in seen:
                continue
            seen.add(k)
            keys.append(k)
    keys.append("notes")
    return keys


def create_issue_link(row):
    link = f"https://github.com/HUPO-PSI/mzSpecLib/issues/new?labels=validation&title={row['id'].replace(' ', '+')}"
    return link


def rule_set_to_worksheet(rule_set, sheet, wrap_fmt, default_fmt):
    rows = list(itertools.chain.from_iterable(map(rule_to_table_rows, rule_set["rules"])))
    headers = infer_headers(rows)
    sheet.write_row(0, 0, headers + ["link"], default_fmt)

    for i, row in enumerate(rows, 1):
        sheet.write_row(i, 0, [row.get(k) for k in headers])
        sheet.write_url(i, len(headers), create_issue_link(row), string="Open Issue")

    for i, h in enumerate(headers):
        if h == "notes":
            sheet.set_column(i, i, 60, cell_format=wrap_fmt)
        elif h == "attribute_name":
            sheet.set_column(i, i, 30, cell_format=default_fmt)
        elif h == "description":
            sheet.set_column(i, i, 60, cell_format=wrap_fmt)
        elif h in ("level", "combination_logic"):
            sheet.set_column(i, i, 10, cell_format=default_fmt)
        else:
            sheet.set_column(i, i, 20, cell_format=default_fmt)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("outpath", help="The name of the file to write")
    parser.add_argument("rule_sets", nargs="+", help="The rule set JSON file paths to convert")
    parser.add_argument("-f", "--format", choices=["xlsx", "csv"])
    args = parser.parse_args()

    rule_set_paths = list(filter(lambda x: not x.endswith("rules-schema.json"), args.rule_sets))

    if args.format == 'xlsx':
        with xlsxwriter.Workbook(args.outpath) as wb:
            wrap_fmt = wb.add_format({"text_wrap": True, "align": "vcenter"})
            default_fmt = wb.add_format({"align": "vcenter"})

            for path in rule_set_paths:
                rule_set = json.load(open(path))
                name = rule_set["name"]
                sheet = wb.add_worksheet(name)
                rule_set_to_worksheet(rule_set, sheet, wrap_fmt, default_fmt)

    elif args.format == 'csv':
        with open(args.outpath, 'wt', newline='') as fh:
            all_rows = []
            for path in rule_set_paths:
                rule_set = json.load(open(path))
                rows = list(itertools.chain.from_iterable(map(rule_to_table_rows, rule_set["rules"])))
                for row in rows:
                    row['rule_set_name'] = rule_set['name']
                    row["link"] = create_issue_link(row)
                all_rows.extend(rows)

            headers = infer_headers(all_rows)
            headers.remove("rule_set_name")
            headers.remove("link")
            headers = ['rule_set_name'] + headers + ['link']
            writer = csv.DictWriter(fh, headers)
            writer.writeheader()
            writer.writerows(all_rows)
            fh.flush()



if __name__ == "__main__":
    main()


from mzspeclib import SpectrumLibrary

# Open a spectrum library
lib = SpectrumLibrary(filename="examples/fetal_brain_tiny.mzlib.txt")
print(lib)

# Get the number of spectra in the library
n_spectra = len(lib)
print(n_spectra)

# Get a specific spectrum from the library
spec = lib.get_spectrum(spectrum_number=3)
print(f"Key={spec.key}; Name={spec.name}; Num Peaks={len(spec.peak_list)}")
print(spec.get_interpretation('1'))

# Loop over a library and count the number of unique peptide analytes
unique_analytes = set()
n_spectra = 0
for spec in lib:
    for analyte in spec.analytes.values():
        unique_analytes.add(str(analyte.peptide))
    n_spectra += 1

print(f"\n{len(unique_analytes)} unique analytes over {n_spectra} spectra")
print(unique_analytes)


#!/usr/bin/env python3

# pragma: no cover
from __future__ import print_function
import sys
def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)

import os
from datetime import datetime

from sqlalchemy import Column, ForeignKey, Integer, Float, String, DateTime, Text, PickleType, LargeBinary
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy import desc
from sqlalchemy import inspect

Base = declarative_base()

debug = False


#### Define the database tables as classes
class SpectrumLibraryIndexAttribute(Base):
  __tablename__ = 'spectrum_library_index_attribute'
  id = Column(Integer, primary_key=True)
  name = Column(String(255), nullable=False)
  value = Column(String(1024), nullable=False)


#### Define the database tables as classes
class SpectrumLibraryIndexRecord(Base):
  __tablename__ = 'spectrum_library_index_record'
  id = Column(Integer, primary_key=True)
  number = Column(Integer, nullable=False)
  offset = Column(Integer, nullable=False)
  name = Column(String(1024), nullable=False)
  peptide_sequence = Column(String(2014), nullable=True)


class SpectrumLibraryIndex:
    """
    SpectrumLibraryIndex - Class for a spectrum library index

    Attributes
    ----------
    columns : array
        Names of the columns in the data matrix

    Methods
    -------
    get_offset - Get the offset for a spectrum in the library based on the spectrum_index or spectrum_name
    find_offsets - Return an array of offsets of spectra that match the input parameters
    create_index - Create a new index for a library
    add_spectrum - Add a spectrum to the index

    """


    #### Constructor
    def __init__(self, library_filename=None, version=None, n_spectra=None, library_datetime=None, columns=None):
        """
        __init__ - SpectrumLibraryIndex constructor

        Parameters
        ----------
        columns : array
            Names of the columns in the data matrix

        """

        self.library_filename = library_filename
        self.version = "0.1"
        self.n_spectra = 0
        self.library_datetime = None
        self.columns = [ 'number', 'offset', 'name', 'peptide_sequence' ]
        self.status = 'closed'
        self.uncommitted_transactions = 0

        if library_filename is None:
            raise Exception('Library filename missing')
        filename = self.library_filename + '.splindex'
        if filename is None:
            raise Exception('Missing library_filename')
        if os.path.exists(filename):
            self.connect()
            self.status = 'OK'
        else:
            self.create_database()
            self.status = 'OK'


    #### Destructor
    def __del__(self):
        if self.library_filename is not None:
            self.disconnect()


    #### Define getter/setter for attribute library_filename
    @property
    def library_filename(self):
        return(self._library_filename)
    @library_filename.setter
    def library_filename(self, library_filename):
        self._library_filename = library_filename

    #### Define getter/setter for attribute version
    @property
    def version(self):
        return(self._version)
    @version.setter
    def version(self, version):
        self._version = version

    #### Define getter/setter for attribute n_spectra
    @property
    def n_spectra(self):
        return(self._n_spectra)
    @n_spectra.setter
    def n_spectra(self, n_spectra):
        self._n_spectra = n_spectra

    #### Define getter/setter for attribute library_datetime
    @property
    def library_datetime(self):
        return(self._library_datetime)
    @library_datetime.setter
    def library_datetime(self, library_datetime):
        self._library_datetime = library_datetime

    #### Define getter/setter for attribute columns
    @property
    def columns(self):
        return(self._columns)
    @columns.setter
    def columns(self, columns):
        self._columns = columns


    #### Define attribute session
    @property
    def session(self) -> str:
        return self._session

    @session.setter
    def session(self, session: str):
        self._session = session


    #### Define attribute engine
    @property
    def engine(self) -> str:
        return self._engine

    @engine.setter
    def engine(self, engine: str):
        self._engine = engine


    #### Delete and create the database. Careful!
    def create_database(self):
        filename = self.library_filename + '.splindex'
        if os.path.exists(filename):
            if debug: eprint(f'INFO: Deleting previous index file {filename}')
            os.remove(filename)
        if debug: eprint(f'INFO: Creating index file {filename}')
        engine = create_engine("sqlite:///"+filename)
        Base.metadata.create_all(engine)

        DBSession = sessionmaker(bind=engine)
        session = DBSession()
        self.session = session
        self.engine = engine

        index_attribute = SpectrumLibraryIndexAttribute( name='version', value=self.version )
        session.add(index_attribute)
        index_attribute = SpectrumLibraryIndexAttribute( name='n_spectra', value=0 )
        session.add(index_attribute)

        session.flush()
        session.commit()


    #### Create and store a database connection
    def connect(self):
        filename = self.library_filename + '.splindex'
        if debug: eprint(f'INFO: Opening index file {filename}')
        engine = create_engine("sqlite:///"+filename)
        DBSession = sessionmaker(bind=engine)
        session = DBSession()
        self.session = session
        self.engine = engine


    #### Destroy the database connection
    def disconnect(self):
        filename = self.library_filename + '.splindex'
        if self.uncommitted_transactions > 0:
            if debug: eprint(f'INFO: Committing open transactions')
            self.session.flush()
            self.session.commit()
            self.uncommitted_transactions = 0

        if debug: eprint(f'INFO: Closing index file {filename}')
        session = self.session
        engine = self.engine
        session.close()
        engine.dispose()
        self.session = None
        self.engine = None


    #### Destroy the database connection
    def commit(self):
        if self.uncommitted_transactions > 0:
            #if debug: eprint(f'INFO: Committing open transactions')
            self.session.flush()
            self.session.commit()
            self.uncommitted_transactions = 0


    def get_offset(self, spectrum_index_number=None, spectrum_name=None):
        """
        get_offset - Get the offset for a spectrum in the library based on the spectrum_index or spectrum_name

        Extended description of function.

        Parameters
        ----------
        spectrum_index : integer
            Index number of the spectrum to select
        spectrum_name : string
            Name of the spectrum to select

        Returns
        -------
        int
            Description of return value
        """

        #### Begin functionality here
        #TODO: Check if SpectrumLibrary is empty or not yet indexed.
        if spectrum_index_number is not None:
            try:
                records = self.session.query(SpectrumLibraryIndexRecord).filter(SpectrumLibraryIndexRecord.number==spectrum_index_number).all()
            except:
                self.session.rollback()
                raise
            if len(records) > 1:
                raise Exception('Too many records')
            return(records[0].offset)

        return()


    def find_offsets(self):
        """
        find_offsets - Return an array of offsets of spectra that match the input parameters

        Extended description of function.

        Parameters
        ----------

        Returns
        -------
        int
            Description of return value
        """

        #### Begin functionality here

        return()


    def create_index(self):
        """
        create_index - Create a new index for a library

        Extended description of function.

        Parameters
        ----------

        Returns
        -------
        int
            Description of return value
        """

        #### Begin functionality here
        if self.session is not None:
            self.disconnect()
        self.create_database()
        return(True)


    def add_spectrum(self, number=None, offset=None, name=None, peptide_sequence=None):
        """
        add_spectrum - Add a spectrum to the index

        Extended description of function.

        Parameters
        ----------
        number : integer
            Index number of the spectrum to add
        offset : integer
            File offset of the spectrum to add
        name : string
            Name of the spectrum to add
        peptide_sequence : string
            Unmodified peptide sequence of the spectrum to add

        Returns
        -------
        int
            Description of return value
        """

        #### Begin functionality here
        #if debug: eprint("INFO: Adding an index entry")
        session = self.session
        index_record = SpectrumLibraryIndexRecord( number=number, offset=offset, name=name, peptide_sequence=peptide_sequence )
        session.add(index_record)
        self.uncommitted_transactions += 1
        if self.uncommitted_transactions >= 5000:
            session.flush()
            session.commit()
            self.uncommitted_transactions = 0

        self.status = 'uncommitted changes'
        return()



#### Example using this class
def example():

    #### Create a new RTXFeedback object
    index = SpectrumLibraryIndex(library_filename='../refData/sigmaups1_consensus_final_true_lib.msp')
    print(index.version)
    return()


#### If this class is run from the command line, perform a short little test to see if it is working correctly
def main():

    #### Run an example
    example()
    return()


if __name__ == "__main__": main()




#!/usr/bin/env python3
# pragma: no cover
from __future__ import print_function
import sys
def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)

import re

class UniversalSpectrumIdentifier(object):

    # usi object takes usiStr an automatically parses it and stores attributes
    # usi objects can still exist even if the usi str is incorrect.
    # it will simply show where the error in the string is

    def __init__(self, usi=None):

        self.is_valid = False
        self.usi = usi
        self.dataset_identifier = None
        self.dataset_subfolder = None
        self.ms_run_name = None
        self.indexFlag = None
        self.index = None
        self.interpretation = None
        self.peptidoform = None
        self.charge = None
        self.provenance_identifier = None
        self.error = 0
        self.error_code = None
        self.error_message = None
        self.warning_message = None

        if usi:
            self.parse(verbose=None)


    # Attributes:
    #   usi
    #   dataset_identifier
    #   dataset_subfolder
    #   ms_run_name
    #   indexFlag
    #   index
    #   interpretation
    #   peptidoform
    #   charge


    #### Set the error state with supplied information
    def set_error(self,error_code,error_message):
        self.error_code = error_code
        self.error_message = error_message
        self.is_valid = False
        #print(f"ERROR: {error_code}: {error_message}")

    # parses USI string
    def parse(self, verbose):
        verboseprint = print if verbose else lambda *a, **k: None
        verboseprint("\nINFO: Parsing USI string '" + self.usi + "'")
        elementOffset = 0
        offset = 0
        if self.usi.startswith("mzspec:"):
            self.usiMzspec = self.usi[len("mzspec:"):]
        else:
            self.set_error("MissingPrefix","USI does not begin with prefix 'mszpec:'")
            return(self)

        # creates list of potential usi attributes
        elements = self.usiMzspec.split(":")
        nElements = len(elements)
        # print(elements)
        # print(nElements)

        # checks if usi has at least 4 colon-separated fields
        if nElements < 4:
            self.set_error("InsufficientComponents","USI string does not have the minimum required 4 colon-separated fields after mzspec:")
            return(self)

        offset = elementOffset

        # dataset_identifier field
        self.dataset_identifier = elements[offset]
        if self.dataset_identifier is None:
            self.set_error("EmptyDatasetIdentifier","Dataset identifier is empty. Not permitted.")
        elif self.dataset_identifier.startswith("PXD") or self.dataset_identifier.startswith("PXL"):
            self.dataset_identifier = elements[offset]
        else:
            self.set_error("UnrecognizedDatasetIdentifier","Dataset identifier unknown. Not permitted.")
            self.error += 1

        elementOffset += 1
        offset = elementOffset
        nextField = elements[offset]
        offsetShift = 0

        #### The old style thing now clashes with the interest in leaving the version out of PXLs
        # empty dataset_subfolder
        #if nextField == '':
        #    self.warning_message = "old style. empty is ok. Empty dataset_subfolder probably."
        #    offsetShift = 1

        offset = elementOffset + offsetShift
        self.ms_run_name = elements[offset]

        if self.ms_run_name:
            verboseprint("MS run equals " + self.ms_run_name)
        else:
            if self.dataset_identifier.startswith("PXL"):
                verboseprint("For PXL identifiers, the version may be blank, okay.")
            else:
                self.set_error("MissingMSRunIdentifier","MS Run identifier empty. Not permitted.")
                self.error += 1

        elementOffset += 1
        offset = elementOffset + offsetShift
        self.indexFlag = elements[offset]
        # print("check " + self.indexFlag)
        # does indexFlag exist?
        if self.indexFlag:
            # is it scan or index
            if self.indexFlag == "scan" or self.indexFlag == "index":
                verboseprint("indexFlag is OK.")
            # is there potentially some weird colon escaping in the msRun name?
            else:
                potentialOffsetShift = offsetShift
                appendStr = ""
                repaired = False

                # fix colon escaping if it exists
                while elementOffset + potentialOffsetShift < nElements:
                    # go until program finds 'scan' or 'index' index flag types
                    if elements[elementOffset + potentialOffsetShift].startswith("scan") or elements[
                        elementOffset + potentialOffsetShift].startswith("index"):
                        self.indexFlag = elements[elementOffset + potentialOffsetShift]
                        self.ms_run_name += appendStr
                        offsetShift = potentialOffsetShift
                        repaired = True
                        break
                    appendStr += ":" + elements[elementOffset + potentialOffsetShift]

                    potentialOffsetShift += 1

                # colon escape fixed and msRun field updated
                if repaired:
                    verboseprint("Unescaped colon in msRun name. Hopefully taken care of. Please fix this")
                    verboseprint("msRun name revised to '{}'".format(self.ms_run_name))

                # no 'scan' or 'index' fields found later. assume broken index flag
                else:
                    self.error += 1
                    self.set_error("InvalidIndexType","Index type invalid. Must be 'scan' or 'index'")

        # no index flag
        else:
            self.error += 1
            self.set_error("MissingIndexType","Index flag empty! Not permitted.")

        elementOffset += 1
        offset = offsetShift + elementOffset

        # index for index flag if flag is valid. useless if index flag is invalid
        self.index = elements[offset]
        if self.index:
            verboseprint("Index is " + self.index)
        else:
            self.set_error("MissingIndex","Index number empty! Not permitted.")
            self.error += 1

        elementOffset += 1
        offset = elementOffset + offsetShift

        # if statement check to see if the USI even has an interpretation field
        if offset < nElements:
            self.interpretation = elements[offset]
            self.peptidoform = ''
            self.charge = ''
            if self.interpretation and self.interpretation != '':
                find = re.match(r"^\s*(.+)\/(\d+)\s*$", self.interpretation)
                # match
                if find:
                    # subfields of interpretation
                    self.peptidoform = find.group(1)
                    self.charge = find.group(2)
                    verboseprint("Interpreted peptidoform = {}, charge = {}".format(self.peptidoform, self.charge))
                else:
                    self.set_error("MissingIndex","Unable to parse interpretation {} as peptidoform/charge".format(self.interpretation))

            else:
                verboseprint("Interpretation field not provided. OK.")

        # provenance identifier
        if offset < nElements:
            self.provenance_identifier = elements[offset]
            verboseprint("Provenance Identifier = ".format(self.provenance_identifier))
        # returns count of errors found in usi. useful for checking if the entire identifier is valid.

        if self.error == 0:
           self.is_valid = True
        # errors found in usi
        else:
            verboseprint("Number of errors: " + str(self.error))
            self.is_valid = False
            verboseprint("ERROR: Invalid USI " + self.usi)

    # prints out USI attributes
    def show(self):
        print("USI: " + str(self.usi))
        print("is_valid: " + str(self.is_valid))
        print("Dataset Identifier: " + str(self.dataset_identifier))
        print("Dataset Subfolder: " + str(self.dataset_subfolder))
        print("MS run name: " + str(self.ms_run_name))
        print("Index flag: " + str(self.indexFlag))
        print("Index: " + str(self.index))
        print("Peptidoform: " + str(self.peptidoform))
        print("Charge: " + str(self.charge))


# If this class is run from the command line, perform a short little test to see if it is working correctly
def run_tests():
    testUSIs = [
        ["valid", "mzspec:PXD002437:00261_A06_P001564_B00E_A00_R1:scan:10951"],
        ["invalid", "PXD002437:00261_A06_P001564_B00E_A00_R1:scan:10951"],
        ["valid", "mzspec:PXD002437:00261_A06_P001564_B00E_A00_R1:scan:10951:PEPT[Phospho]IDELVISK/2"],
        ["invalid", "mzspec:PASS002437:00261_A06_P001564_B00E_A00_R1:scan:10951:PEPT[Phospho]IDELVISK/2"],
        ["invalid", "mzspec"],
        ["invalid", "mzspec:"],
        ["invalid", "mzspec:PXD001234"],
        ["invalid", "mzspec:PXD001234:00261_A06_P001564_B00E_A00_R1:scan"],
        ["valid", "mzspec:PXD001234:00261_A06_P001564_B00E_A00_R1:index:10951"],
        ["valid", "mzspec:PXD002437:00261_A06_P001564_B00E_A00_R1:scan:10951:PEPT[Phospho]IDELVISK/2"],
        ["valid", "mzspec:PXD002437:00261_A06_P001564_B00E_A00_R1:scan:10951:PEPT[+79]IDELVISK/2"],
        ["valid", "mzspec:PXD001234:Dilution1:4:scan:10951"],
        ["valid", "mzspec:PXD002437:00261_A06_P001564_B00E_A00_R1:test1:scan:10951:PEPT[Phospho]IDELVISK/2"],
        ["valid", "mzspec:PXD002437:00261_A06_P001564_B00E_A00_R1\\:test1:scan:10951:PEPT[Phospho]IDELVISK/2"],
    ]
    testUSIsValid = []

    # Loop over each test USI, parse it, and determine if it is valid or not, and print the index number
    print("Testing example USIs:")
    for usiSet in testUSIs:
        expectedStatus = usiSet[0]
        usiStr = usiSet[1]

        # Create a new UniversalSpectrumIdentifier object
        # made the USI object itself take a string so that parse does not need to be called explicitly
        usi = UniversalSpectrumIdentifier(usiStr)
        expected_validity = True
        if expectedStatus == 'invalid':
            expected_validity = False
        else:
            expectedStatus = 'valid  '

        status = 'PASS'
        if usi.is_valid is not expected_validity:
            status = 'FAIL'

        response = usi.is_valid
        testUSIsValid.append(response)
        print(f"{status}\texpected {expectedStatus}\t{usiStr}")
    # check to see if parsing is correct
    #print(testUSIsValid)


#### A very simple example of using this class
def example():
    usiStr = "mzspec:PXD002437:00261_A06_P001564_B00E_A00_R1:scan:10951"
    usiStr = "mzspec:PXL000001:01-09-2015:index:500"
    usiStr = "mzspec:PXL000001::index:500"
    usi = UniversalSpectrumIdentifier(usiStr)
    #usi.parse(verbose=1)
    usi.show()


#### If class is run directly
def main():
    #example()
    run_tests()


if __name__ == "__main__": main()
